# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 Lorenz Hüdepohl <dev@stellardeath.org>             *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import Path.Base.Generator.helix as helix
from PathScripts.PathUtils import fmt
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import Part
import Path
import Path.Base.FeedRate as PathFeedRate
import Path.Op.Base as PathOp
import Path.Op.CircularHoleBase as PathCircularHoleBase


__title__ = "CAM Helix Operation"
__author__ = "Lorenz Hüdepohl"
__url__ = "https://www.freecad.org"
__doc__ = "Class and implementation of Helix Drill operation"
__contributors__ = "russ4262 (Russell Johnson)"
__created__ = "2016"
__scriptVersion__ = "1b testing"
__lastModified__ = "2019-07-12 09:50 CST"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


def _caclulatePathDirection(mode, side):
    """Calculates the path direction from cut mode and cut side"""
    # NB: at the time of writing, we need py3.8 compat, thus not using py3.10 pattern machting
    if mode == "Conventional" and side == "Inside":
        return "CW"
    elif mode == "Conventional" and side == "Outside":
        return "CCW"
    elif mode == "Climb" and side == "Inside":
        return "CCW"
    elif mode == "Climb" and side == "Outside":
        return "CW"
    else:
        raise ValueError(f"No mapping for '{mode}'/'{side}'")


def _caclulateCutMode(direction, side):
    """Calculates the cut mode from path direction and cut side"""
    # NB: at the time of writing, we need py3.8 compat, thus not using py3.10 pattern machting
    if direction == "CW" and side == "Inside":
        return "Conventional"
    elif direction == "CW" and side == "Outside":
        return "Climb"
    elif direction == "CCW" and side == "Inside":
        return "Climb"
    elif direction == "CCW" and side == "Outside":
        return "Conventional"
    else:
        raise ValueError(f"No mapping for '{direction}'/'{side}'")


class ObjectHelix(PathCircularHoleBase.ObjectOp):
    """Proxy class for Helix operations."""

    @classmethod
    def helixOpPropertyEnumerations(self, dataType="data"):
        """helixOpPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        # Enumeration lists for App::PropertyEnumeration properties
        enums = {
            "Direction": [
                (translate("CAM_Helix", "CW"), "CW"),
                (translate("CAM_Helix", "CCW"), "CCW"),
            ],  # this is the direction that the profile runs
            "StartSide": [
                (translate("PathProfile", "Outside"), "Outside"),
                (translate("PathProfile", "Inside"), "Inside"),
            ],  # side of profile that cutter is on in relation to direction of profile
            "CutMode": [
                (translate("CAM_Helix", "Climb"), "Climb"),
                (translate("CAM_Helix", "Conventional"), "Conventional"),
            ],  # whether the tool "rolls" with or against the feed direction along the profile
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def circularHoleFeatures(self, obj):
        """circularHoleFeatures(obj) ... enable features supported by Helix."""
        return PathOp.FeatureStepDown | PathOp.FeatureBaseEdges | PathOp.FeatureBaseFaces

    def initCircularHoleOperation(self, obj):
        """initCircularHoleOperation(obj) ... create helix specific properties."""
        obj.addProperty(
            "App::PropertyEnumeration",
            "Direction",
            "Helix Drill",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The direction of the circular cuts, ClockWise (CW), or CounterClockWise (CCW)",
            ),
        )
        obj.setEditorMode("Direction", ["ReadOnly", "Hidden"])
        obj.setPropertyStatus("Direction", ["ReadOnly", "Output"])

        obj.addProperty(
            "App::PropertyEnumeration",
            "StartSide",
            "Helix Drill",
            QT_TRANSLATE_NOOP("App::Property", "Start cutting from the inside or outside"),
        )

        # TODO: revise property description once v1.0 release string freeze is lifted
        obj.addProperty(
            "App::PropertyEnumeration",
            "CutMode",
            "Helix Drill",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The direction of the circular cuts, ClockWise (Climb), or CounterClockWise (Conventional)",
            ),
        )

        obj.addProperty(
            "App::PropertyPercent",
            "StepOver",
            "Helix Drill",
            QT_TRANSLATE_NOOP(
                "App::Property", "Percent of cutter diameter to step over on each pass"
            ),
        )
        obj.addProperty(
            "App::PropertyLength",
            "StartRadius",
            "Helix Drill",
            QT_TRANSLATE_NOOP("App::Property", "Starting Radius"),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "OffsetExtra",
            "Helix Drill",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra value to stay away from final profile- good for roughing toolpath",
            ),
        )

        ENUMS = self.helixOpPropertyEnumerations()
        for n in ENUMS:
            setattr(obj, n[0], n[1])
        obj.StepOver = 50

    def opOnDocumentRestored(self, obj):
        if not hasattr(obj, "StartRadius"):
            obj.addProperty(
                "App::PropertyLength",
                "StartRadius",
                "Helix Drill",
                QT_TRANSLATE_NOOP("App::Property", "Starting Radius"),
            )

        if not hasattr(obj, "OffsetExtra"):
            obj.addProperty(
                "App::PropertyDistance",
                "OffsetExtra",
                "Helix Drill",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Extra value to stay away from final profile- good for roughing toolpath",
                ),
            )

        if not hasattr(obj, "CutMode"):
            # TODO: consolidate the duplicate definitions from opOnDocumentRestored and
            # initCircularHoleOperation once back on the main line
            obj.addProperty(
                "App::PropertyEnumeration",
                "CutMode",
                "Helix Drill",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The direction of the circular cuts, ClockWise (Climb), or CounterClockWise (Conventional)",
                ),
            )
            obj.CutMode = ["Climb", "Conventional"]
            if obj.Direction in ["Climb", "Conventional"]:
                # For some month, late in the v1.0 release cycle, we had the cut mode assigned
                # to the direction (see PR#14364). Let's fix files created in this time as well.
                new_dir = "CW" if obj.Direction == "Climb" else "CCW"
                obj.Direction = ["CW", "CCW"]
                obj.Direction = new_dir
            obj.CutMode = _caclulateCutMode(obj.Direction, obj.StartSide)
            obj.setEditorMode("Direction", ["ReadOnly", "Hidden"])
            obj.setPropertyStatus("Direction", ["ReadOnly", "Output"])

    def circularHoleExecute(self, obj, holes):
        """circularHoleExecute(obj, holes) ... generate helix commands for each hole in holes"""
        Path.Log.track()
        obj.Direction = _caclulatePathDirection(obj.CutMode, obj.StartSide)

        self.commandlist.append(Path.Command("(helix cut operation)"))

        self.commandlist.append(Path.Command("G0", {"Z": obj.ClearanceHeight.Value}))

        tool = obj.ToolController.Tool
        tooldiamter = tool.Diameter.Value if hasattr(tool.Diameter, "Value") else tool.Diameter

        args = {
            "edge": None,
            "hole_radius": None,
            "step_down": obj.StepDown.Value,
            "step_over": obj.StepOver / 100,
            "tool_diameter": tooldiamter,
            "inner_radius": obj.StartRadius.Value + obj.OffsetExtra.Value,
            "direction": obj.Direction,
            "startAt": obj.StartSide,
        }

        for hole in holes:
            args["hole_radius"] = (hole["r"] / 2) - (obj.OffsetExtra.Value)
            startPoint = FreeCAD.Vector(hole["x"], hole["y"], obj.StartDepth.Value)
            endPoint = FreeCAD.Vector(hole["x"], hole["y"], obj.FinalDepth.Value)
            args["edge"] = Part.makeLine(startPoint, endPoint)

            # move to starting position
            self.commandlist.append(Path.Command("G0", {"Z": obj.ClearanceHeight.Value}))
            self.commandlist.append(
                Path.Command(
                    "G0",
                    {
                        "X": startPoint.x,
                        "Y": startPoint.y,
                        "Z": obj.ClearanceHeight.Value,
                    },
                )
            )
            self.commandlist.append(
                Path.Command("G0", {"X": startPoint.x, "Y": startPoint.y, "Z": startPoint.z})
            )

            results = helix.generate(**args)

            for command in results:
                self.commandlist.append(command)

        PathFeedRate.setFeedRate(self.commandlist, obj.ToolController)


def SetupProperties():
    """Returns property names for which the "Setup Sheet" should provide defaults."""
    setup = []
    setup.append("CutMode")
    setup.append("StartSide")
    setup.append("StepOver")
    setup.append("StartRadius")
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Helix operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectHelix(obj, name, parentJob)
    if obj.Proxy:
        obj.Proxy.findAllHoles(obj)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019, 2020 Carlo Pavan <carlopav@gmail.com>             *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides support functions to edit Arch objects."""
## @package gui_edit_arch_objects
# \ingroup draftguitools
# \brief Provides support functions to edit Arch objects.

__title__ = "FreeCAD Draft Edit Tool"
__author__ = "Carlo Pavan"
__url__ = "https://www.freecad.org"

## \addtogroup draftguitools
# @{


class GuiTools:
    """Base class for object editing tools"""

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        """Return to Draft_Edit a list of vectors for the given object.
        Remember to use object local coordinates system.

        Parameters:
        obj: the object
        """
        pass

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        """Update the object from modified Draft_Edit point.
        No need to recompute at the end.

        Parameters:
        obj: the object
        node_idx: number of the edited node
        v: target vector of the node in object local coordinates system
        alt_edit_mode: alternative edit mode to perform different operations
                       (usually can be selected from the Draft_Edit context menu)
                       default = 0
        """
        pass

    def get_edit_point_context_menu(self, edit_command, obj, node_idx):
        """Get the context menu associated to edit points (user is over an editpoint)

        Return a list of tuples containing menu labels and associated functions:
            return [
                ("action label", lambda: self.handle_action_label(edit_command, obj, node_idx)),
            ]

        Parameters:
        edit_command: running Draft_Edit command
        obj: the edited object
        node_idx: number of the edited node
        """
        pass

    def get_edit_obj_context_menu(self, edit_command, obj, position):
        """Get the context menu associated to edited object (user is over the object)

        Return a list of tuples containing menu labels and associated functions:
            return [
                ("action label", lambda: self.handle_action_label(edit_command, obj, position)),
            ]

        Parameters:
        edit_command: running Draft_Edit command
        obj: the edited object
        position: position of the cursor on the screen (x, y)
        """
        pass

    def get_object_style(self, obj):
        pass

    def set_object_editing_style(self, obj):
        pass

    def restore_object_style(self, obj, modes):
        pass

    def init_preview_object(self, obj):
        pass

    def update_preview_object(self, edit_command, obj, node_idx, v):
        pass


## @}



from Base.Metadata import export, no_args
from Base.PyObjectBase import PyObjectBase
from typing import List, Tuple, TypeAlias


MeasureType: TypeAlias = object

@export(
    Constructor=False,
    Delete=True,
    )
class MeasureManager(PyObjectBase):
    """
    MeasureManager class.

    The MeasureManager handles measure types and geometry handler across FreeCAD.

    Author: David Friedli (david@friedli-be.ch)
    Licence: LGPL
    DeveloperDocu: MeasureManager
    """

    @staticmethod
    def addMeasureType(id: str, label: str, measureType: MeasureType) -> None:
        """
        addMeasureType(id, label, measureType) -> None

        Add a new measure type.

        id : str
            Unique identifier of the measure type.
        label : str
            Name of the module.
        measureType : Measure.MeasureBasePython
            The actual measure type.
        """
        ...

    @staticmethod
    @no_args
    def getMeasureTypes() -> List[Tuple[str, str, MeasureType]]:
        """
        getMeasureTypes() -> List[(id, label, pythonMeasureType)]

        Returns a list of all registered measure types.
        """
        ...



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import Arch
import Draft
import FreeCAD as App
from bimtests import TestArchBase


class TestArchFrame(TestArchBase.TestArchBase):

    def test_makeFrame(self):
        """Test the makeFrame function."""

        obj = Arch.makeFrame(None, None)
        self.assertIsNotNone(obj, "makeFrame failed to create an object")
        self.assertEqual(obj.Label, "Frame", "Incorrect default label for Frame")

    def testFrame(self):
        l = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(-2, 0, 0))
        p = Draft.makeRectangle(length=0.5, height=0.5)
        f = Arch.makeFrame(l, p)
        self.assertTrue(f, "Arch Frame failed")



# ***************************************************************************
# *   Copyright (c) 2015 Qingfeng Xia <qingfeng.xia()eng.ox.ac.uk>          *
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2024 PMcB                                               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD result mechanical task panel"
__author__ = "Qingfeng Xia, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package view_result_mechanical
#  \ingroup FEM
#  \brief task panel for mechanical ResultObjectPython

import CreateLabels
import inspect, sys

try:
    import matplotlib

    matplotlib.use("Qt5Agg")
except Exception:
    print("Failed to set matplotlib backend to Qt5Agg")

import matplotlib.pyplot as plt
import numpy as np
import time, math

from PySide import QtCore
from PySide import QtGui
from PySide.QtCore import Qt
from PySide.QtGui import QApplication

import FreeCAD
import FreeCADGui

import femresult.resulttools as resulttools

translate = FreeCAD.Qt.translate


class _TaskPanel:
    """
    The task panel for the post-processing
    """

    def __init__(self, obj):
        self.result_obj = obj
        self.mesh_obj = self.result_obj.Mesh
        # task panel should be started by use of setEdit of view provider
        # in view provider checks: Mesh, active analysis and
        # if Mesh and result are in active analysis
        # activate the result mesh object
        self.mesh_obj.ViewObject.show()

        ui_path = FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/"
        self.result_widget = FreeCADGui.PySideUic.loadUi(ui_path + "ResultShow.ui")
        self.info_widget = FreeCADGui.PySideUic.loadUi(ui_path + "ResultHints.ui")
        self.form = [self.result_widget, self.info_widget]
        self.results_name = "No Contour Data"
        self.animate_inc = 1
        self.startAnimate = False
        self.animateText = []

        self.fem_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/General")
        self.restore_result_settings_in_dialog = self.fem_prefs.GetBool("RestoreResultDialog", True)

        # Connect Signals and Slots
        # result type radio buttons
        # TODO: move to combo box, to be independent from result types and result types count
        QtCore.QObject.connect(
            self.result_widget.rb_none,
            QtCore.SIGNAL("toggled(bool)"),
            self.none_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_abs_displacement,
            QtCore.SIGNAL("toggled(bool)"),
            self.abs_displacement_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_x_displacement,
            QtCore.SIGNAL("toggled(bool)"),
            self.x_displacement_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_y_displacement,
            QtCore.SIGNAL("toggled(bool)"),
            self.y_displacement_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_z_displacement,
            QtCore.SIGNAL("toggled(bool)"),
            self.z_displacement_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_temperature,
            QtCore.SIGNAL("toggled(bool)"),
            self.temperature_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_vm_stress,
            QtCore.SIGNAL("toggled(bool)"),
            self.vm_stress_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_maxprin,
            QtCore.SIGNAL("toggled(bool)"),
            self.max_prin_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_minprin,
            QtCore.SIGNAL("toggled(bool)"),
            self.min_prin_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_max_shear_stress,
            QtCore.SIGNAL("toggled(bool)"),
            self.max_shear_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_massflowrate,
            QtCore.SIGNAL("toggled(bool)"),
            self.massflowrate_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_networkpressure,
            QtCore.SIGNAL("toggled(bool)"),
            self.networkpressure_selected,
        )
        QtCore.QObject.connect(
            self.result_widget.rb_peeq,
            QtCore.SIGNAL("toggled(bool)"),
            self.peeq_selected,
        )

        # stats
        self.result_widget.show_histogram.clicked.connect(self.show_histogram_clicked)
        # animate
        QtCore.QObject.connect(
            self.result_widget.hsb_displacement_factor,
            QtCore.SIGNAL("valueChanged(int)"),
            lambda dummy="", name="scale": self.value_changed(self, dummy, name),
        )
        QtCore.QObject.connect(
            self.result_widget.sb_displacement_factor,
            QtCore.SIGNAL("valueChanged(double)"),
            lambda dummy="", name="factor": self.value_changed(self, dummy, name),
        )
        QtCore.QObject.connect(
            self.result_widget.startButton,
            QtCore.SIGNAL("clicked()"),
            lambda dummy="", name="startButton": self.value_changed(self, dummy, name),
        )

        # displacement
        QtCore.QObject.connect(
            self.result_widget.cb_show_displacement,
            QtCore.SIGNAL("clicked(bool)"),
            self.show_displacement,
        )
        QtCore.QObject.connect(
            self.result_widget.hsb_displacement_factor,
            QtCore.SIGNAL("valueChanged(int)"),
            self.hsb_disp_factor_changed,
        )
        self.result_widget.hsb_displacement_factor.sliderReleased.connect(self.set_focus_sb_disp)

        self.result_widget.sb_displacement_factor.valueChanged.connect(self.sb_disp_factor_changed)
        self.result_widget.sb_displacement_factor_max.valueChanged.connect(
            self.sb_disp_factor_max_changed
        )

        # user defined equation
        self.result_widget.user_def_eq.textChanged.connect(self.user_defined_text)
        QtCore.QObject.connect(
            self.result_widget.calculate, QtCore.SIGNAL("clicked()"), self.calculate
        )

        # Since, the slider is integer based keep it always between 0 to 100
        # and scale the value from the inputbox
        self.result_widget.hsb_displacement_factor.setRange(0, 100)

        self.update()
        if self.restore_result_settings_in_dialog:
            self.restore_result_dialog()
        else:
            self.restore_initial_result_dialog()
            # initialize scale factor for show displacement
            scale_factor = get_displacement_scale_factor(self.result_obj)
            self.result_widget.sb_displacement_factor_max.setValue(10.0 * scale_factor)
            self.result_widget.sb_displacement_factor.setValue(scale_factor)

    def restore_result_dialog(self):
        try:
            rt = FreeCAD.FEM_dialog["results_type"]
            if rt == "None":
                self.result_widget.rb_none.setChecked(True)
                self.none_selected(True)
            elif rt == "Uabs":
                self.result_widget.rb_abs_displacement.setChecked(True)
                self.abs_displacement_selected(True)
            elif rt == "U1":
                self.result_widget.rb_x_displacement.setChecked(True)
                self.x_displacement_selected(True)
            elif rt == "U2":
                self.result_widget.rb_y_displacement.setChecked(True)
                self.y_displacement_selected(True)
            elif rt == "U3":
                self.result_widget.rb_z_displacement.setChecked(True)
                self.z_displacement_selected(True)
            elif rt == "Temp":
                self.result_widget.rb_temperature.setChecked(True)
                self.temperature_selected(True)
            elif rt == "Sabs":
                self.result_widget.rb_vm_stress.setChecked(True)
                self.vm_stress_selected(True)
            elif rt == "MaxPrin":
                self.result_widget.rb_maxprin.setChecked(True)
                self.max_prin_selected(True)
            elif rt == "MinPrin":
                self.result_widget.rb_minprin.setChecked(True)
                self.min_prin_selected(True)
            elif rt == "MaxShear":
                self.result_widget.rb_max_shear_stress.setChecked(True)
                self.max_shear_selected(True)
            elif rt == "MFlow":
                self.result_widget.rb_massflowrate.setChecked(True)
                self.massflowrate_selected(True)
            elif rt == "NPress":
                self.result_widget.rb_networkpressure.setChecked(True)
                self.networkpressure_selected(True)
            elif rt == "Peeq":
                self.result_widget.rb_peeq.setChecked(True)
                self.peeq_selected(True)

            sd = FreeCAD.FEM_dialog["show_disp"]
            self.result_widget.cb_show_displacement.setChecked(sd)
            self.show_displacement(sd)

            df = FreeCAD.FEM_dialog["disp_factor"]
            dfm = FreeCAD.FEM_dialog["disp_factor_max"]
            self.result_widget.sb_displacement_factor_max.setValue(dfm)
            self.result_widget.sb_displacement_factor.setValue(df)
            self.result_widget.hsb_displacement_factor.setValue(self.convert_to_slider_value(dfm))
            # animate
            self.startAnimate = False
            if FreeCAD.FEM_dialog["animate"][0] != -1:
                self.result_widget.steps.setValue(FreeCAD.FEM_dialog["animate"][0])
                self.result_widget.loops.setValue(FreeCAD.FEM_dialog["animate"][1])
                self.result_widget.framerate.setValue(FreeCAD.FEM_dialog["animate"][2])
        except Exception:
            self.restore_initial_result_dialog()

    def restore_initial_result_dialog(self):
        # initialize FreeCAD.FEM_dialog and set standard values
        # the FEM result mechanical task panel restore values
        # are saved in a dictionary which is an attribute of FreeCAD
        # the name is FEM_dialog
        # in python console after result task panel has been opened once
        # FreeCAD.FEM_dialog or FreeCAD.__dir__()
        # This is not smart at all IMHO (Bernd)
        # It was added with commit 3a7772d
        # https://github.com/FreeCAD/FreeCAD/commit/3a7772d
        FreeCAD.FEM_dialog = {
            "results_type": "None",
            "show_disp": True,  # False,
            "disp_factor": 5.0,
            "disp_factor_max": 100.0,
            "animate": [-1, -1, -1, -1],  # steps, loops, rate, indicator (not used)
        }
        self.result_widget.sb_displacement_factor_max.setValue(100.0)  # init non standard values

    def getStandardButtons(self):
        return QtGui.QDialogButtonBox.Close

    def get_result_stats(self, type_name):
        return resulttools.get_stats(self.result_obj, type_name)

    def none_selected(self, state):
        self.set_label(self.result_obj.Label, "No Contours")
        FreeCAD.FEM_dialog["results_type"] = "None"
        self.set_result_stats("mm", 0.0, 0.0)
        self.reset_mesh_color()
        if len(plt.get_fignums()) > 0:
            plt.close()

    # if an analysis has different result types and one has
    # stress and the other not the restore result dialog
    # could trigger stress selected for a result object
    # which has not stress
    # see https://forum.freecad.org/viewtopic.php?f=18&t=39162
    # check if the results len is not 0 on any selected method

    def abs_displacement_selected(self, state):
        if len(self.result_obj.DisplacementLengths) > 0:
            self.result_selected(
                "Uabs",
                self.result_obj.DisplacementLengths,
                "mm",
                translate("FEM", "Displacement Magnitude"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def x_displacement_selected(self, state):
        if len(self.result_obj.DisplacementVectors) > 0:
            res_disp_u1 = self.get_scalar_disp_list(self.result_obj.DisplacementVectors, 0)
            self.result_selected("U1", res_disp_u1, "mm", translate("FEM", "Displacement X"))
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def y_displacement_selected(self, state):
        if len(self.result_obj.DisplacementVectors) > 0:
            res_disp_u2 = self.get_scalar_disp_list(self.result_obj.DisplacementVectors, 1)
            self.result_selected("U2", res_disp_u2, "mm", translate("FEM", "Displacement Y"))
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def z_displacement_selected(self, state):
        if len(self.result_obj.DisplacementVectors) > 0:
            res_disp_u3 = self.get_scalar_disp_list(self.result_obj.DisplacementVectors, 2)
            self.result_selected("U3", res_disp_u3, "mm", translate("FEM", "Displacement Z"))
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def vm_stress_selected(self, state):
        if len(self.result_obj.vonMises) > 0:
            self.result_selected(
                "Sabs",
                self.result_obj.vonMises,
                "MPa",
                translate("FEM", "von Mises Stress"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def max_shear_selected(self, state):
        if len(self.result_obj.MaxShear) > 0:
            self.result_selected(
                "MaxShear",
                self.result_obj.MaxShear,
                "MPa",
                translate("FEM", "Max Shear Stress"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def max_prin_selected(self, state):
        if len(self.result_obj.PrincipalMax) > 0:
            self.result_selected(
                "MaxPrin",
                self.result_obj.PrincipalMax,
                "MPa",
                translate("FEM", "Max Principal Stress"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def temperature_selected(self, state):
        if len(self.result_obj.Temperature) > 0:
            self.result_selected(
                "Temp",
                self.result_obj.Temperature,
                "K",
                translate("FEM", "Temperature"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def massflowrate_selected(self, state):
        if len(self.result_obj.MassFlowRate) > 0:
            self.result_selected(
                "MFlow",
                self.result_obj.MassFlowRate,
                "kg/s",
                translate("FEM", "Mass Flow Rate"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def networkpressure_selected(self, state):
        if len(self.result_obj.NetworkPressure) > 0:
            self.result_selected(
                "NPress",
                self.result_obj.NetworkPressure,
                "MPa",
                translate("FEM", "Network Pressure"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def min_prin_selected(self, state):
        if len(self.result_obj.PrincipalMin) > 0:
            self.result_selected(
                "MinPrin",
                self.result_obj.PrincipalMin,
                "MPa",
                translate("FEM", "Min Principal Stress"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def peeq_selected(self, state):
        if len(self.result_obj.Peeq) > 0:
            self.result_selected(
                "Peeq",
                self.result_obj.Peeq,
                "",
                translate("FEM", "Equivalent Plastic Strain"),
            )
        else:
            self.result_widget.rb_none.setChecked(True)
            self.none_selected(True)

    def show_histogram_clicked(self):
        if len(plt.get_fignums()) > 0:
            plt.show()
        else:
            # if the plot was closed and subsequently the Histogram button was pressed again
            # we have valid settings, but must restore the dialog to refill the plot content
            # see https://github.com/FreeCAD/FreeCAD/issues/6975
            if FreeCAD.FEM_dialog["results_type"] != "None":
                self.restore_result_dialog()
            if len(plt.get_fignums()) > 0:
                plt.show()
            else:
                QtGui.QMessageBox.information(
                    None,
                    self.result_obj.Label + " - " + translate("FEM", "Information"),
                    translate(
                        "FEM",
                        "No histogram available.\nPlease select a result type first.",
                    ),
                )

    def user_defined_text(self, equation):
        FreeCAD.FEM_dialog["results_type"] = "user"
        self.result_widget.user_def_eq.toPlainText()

    def calculate(self):
        # Convert existing result values to numpy array
        # scalars
        P1 = np.array(self.result_obj.PrincipalMax)
        P2 = np.array(self.result_obj.PrincipalMed)
        P3 = np.array(self.result_obj.PrincipalMin)
        MS = np.array(self.result_obj.MaxShear)
        vM = np.array(self.result_obj.vonMises)
        Peeq = np.array(self.result_obj.Peeq)
        T = np.array(self.result_obj.Temperature)
        MF = np.array(self.result_obj.MassFlowRate)
        NP = np.array(self.result_obj.NetworkPressure)
        sxx = np.array(self.result_obj.NodeStressXX)
        syy = np.array(self.result_obj.NodeStressYY)
        szz = np.array(self.result_obj.NodeStressZZ)
        sxy = np.array(self.result_obj.NodeStressXY)
        sxz = np.array(self.result_obj.NodeStressXZ)
        syz = np.array(self.result_obj.NodeStressYZ)
        exx = np.array(self.result_obj.NodeStrainXX)
        eyy = np.array(self.result_obj.NodeStrainYY)
        ezz = np.array(self.result_obj.NodeStrainZZ)
        exy = np.array(self.result_obj.NodeStrainXY)
        exz = np.array(self.result_obj.NodeStrainXZ)
        eyz = np.array(self.result_obj.NodeStrainYZ)
        rx = np.array(self.result_obj.ReinforcementRatio_x)
        ry = np.array(self.result_obj.ReinforcementRatio_y)
        rz = np.array(self.result_obj.ReinforcementRatio_z)
        mc = np.array(self.result_obj.MohrCoulomb)
        # vectors
        dispvectors = np.array(self.result_obj.DisplacementVectors)
        x = np.array(dispvectors[:, 0])
        y = np.array(dispvectors[:, 1])
        z = np.array(dispvectors[:, 2])
        s1x, s1y, s1z = np.array([]), np.array([]), np.array([])
        s2x, s2y, s2z = np.array([]), np.array([]), np.array([])
        s3x, s3y, s3z = np.array([]), np.array([]), np.array([])
        # If PSxVector is empty all UserDefined equation does not work
        if self.result_obj.PS1Vector:
            ps1vector = np.array(self.result_obj.PS1Vector)
            s1x = np.array(ps1vector[:, 0])
            s1y = np.array(ps1vector[:, 1])
            s1z = np.array(ps1vector[:, 2])
        if self.result_obj.PS2Vector:
            ps2vector = np.array(self.result_obj.PS2Vector)
            s2x = np.array(ps2vector[:, 0])
            s2y = np.array(ps2vector[:, 1])
            s2z = np.array(ps2vector[:, 2])
        if self.result_obj.PS3Vector:
            ps3vector = np.array(self.result_obj.PS3Vector)
            s3x = np.array(ps3vector[:, 0])
            s3y = np.array(ps3vector[:, 1])
            s3z = np.array(ps3vector[:, 2])

        FreeCAD.FEM_dialog["results_type"] = "None"
        self.update()
        self.restore_result_dialog()
        userdefined_eq = self.result_widget.user_def_eq.toPlainText()  # Get equation to be used
        self.results_name = "User Defined: " + userdefined_eq

        # https://forum.freecad.org/viewtopic.php?f=18&t=42425&start=10#p368774 ff
        # https://github.com/FreeCAD/FreeCAD/pull/3020
        from ply import lex
        from ply import yacc
        import femtools.tokrules as tokrules

        identifiers = [
            "x",
            "y",
            "z",
            "T",
            "vM",
            "Peeq",
            "P1",
            "P2",
            "P3",
            "sxx",
            "syy",
            "szz",
            "sxy",
            "sxz",
            "syz",
            "exx",
            "eyy",
            "ezz",
            "exy",
            "exz",
            "eyz",
            "MS",
            "MF",
            "NP",
            "rx",
            "ry",
            "rz",
            "mc",
            "s1x",
            "s1y",
            "s1z",
            "s2x",
            "s2y",
            "s2z",
            "s3x",
            "s3y",
            "s3z",
        ]
        tokrules.names = {}
        for i in identifiers:
            tokrules.names[i] = locals()[i]

        lexer = lex.lex(module=tokrules)
        yacc.parse(input=f"UserDefinedFormula={userdefined_eq}", lexer=lexer)
        UserDefinedFormula = tokrules.names["UserDefinedFormula"].tolist()
        tokrules.names = {}
        # UserDefinedFormula = eval(userdefined_eq).tolist()

        if UserDefinedFormula:
            self.result_obj.UserDefined = UserDefinedFormula
            minm = min(UserDefinedFormula)
            maxm = max(UserDefinedFormula)
            self.update_colors_stats(UserDefinedFormula, "", minm, maxm)

        # finally we must recompute the result_obj
        self.result_obj.Document.recompute()

    def get_scalar_disp_list(self, vector_list, axis):
        # list is needed, as zib-object is not subscriptable in py3
        d = list(zip(*self.result_obj.DisplacementVectors))
        scalar_list = list(d[axis])
        return scalar_list

    def result_selected(self, res_type, res_values, res_unit, res_title):
        self.results_name = res_title
        FreeCAD.FEM_dialog["results_type"] = res_type
        (minm, maxm) = self.get_result_stats(res_type)
        self.update_colors_stats(res_values, res_unit, minm, maxm)

        if len(plt.get_fignums()) > 0:
            plt.close()
        plt.ioff()  # disable interactive mode so we have full control when plot is shown
        plt.figure(res_title)
        plt.hist(res_values, bins=50, alpha=0.5, facecolor="blue")
        plt.xlabel(res_unit)
        plt.title(translate("FEM", "Histogram of {}").format(res_title))
        plt.ylabel(translate("FEM", "Nodes"))
        plt.grid(True)
        fig_manager = plt.get_current_fig_manager()
        # Lines below tells Qt that plot widget/dialog should be kept on top of its parent,
        # its parent being defined as FC main window
        # "Tool" type also is non modal, and dialog doesn't add a tab in the OS task bar
        # See Qt::WindowFlags for more details
        fig_manager.window.setParent(FreeCADGui.getMainWindow())
        fig_manager.window.setWindowFlag(QtCore.Qt.Tool)

    def update_colors_stats(self, res_values, res_unit, minm, maxm):
        self.set_label(self.result_obj.Label, self.results_name)
        QApplication.setOverrideCursor(Qt.WaitCursor)
        if self.suitable_results:
            self.mesh_obj.ViewObject.setNodeColorByScalars(self.result_obj.NodeNumbers, res_values)
        self.set_result_stats(res_unit, minm, maxm)
        QtGui.QApplication.restoreOverrideCursor()

    def set_result_stats(self, unit, minm, maxm):
        self.result_widget.le_min.setProperty("unit", unit)
        self.result_widget.le_min.setProperty("rawText", f"{minm:.6} {unit}")
        self.result_widget.le_max.setProperty("unit", unit)
        self.result_widget.le_max.setProperty("rawText", f"{maxm:.6} {unit}")

    def update_displacement(self, factor=None):
        if factor is None:
            if FreeCAD.FEM_dialog["show_disp"]:
                factor = self.result_widget.sb_displacement_factor.value()
            else:
                factor = 0.0
        self.mesh_obj.ViewObject.applyDisplacement(factor)

    def show_displacement(self, checked):
        # Enable or disable the input widgets in the displacement group
        self.result_widget.sb_displacement_factor.setEnabled(checked)
        self.result_widget.sb_displacement_factor_max.setEnabled(checked)
        self.result_widget.hsb_displacement_factor.setEnabled(checked)

        QApplication.setOverrideCursor(Qt.WaitCursor)
        FreeCAD.FEM_dialog["show_disp"] = checked
        if "result_obj" in FreeCAD.FEM_dialog:
            if FreeCAD.FEM_dialog["result_obj"] != self.result_obj:
                self.update_displacement()
        FreeCAD.FEM_dialog["result_obj"] = self.result_obj
        if self.suitable_results:
            self.mesh_obj.ViewObject.setNodeDisplacementByVectors(
                self.result_obj.NodeNumbers, self.result_obj.DisplacementVectors
            )
        self.update_displacement()
        QtGui.QApplication.restoreOverrideCursor()

    def convert_to_slider_value(self, value: float):
        # The slider always goes between 0 to 100 in steps of 1
        if value == 0.0:
            return 0.0

        max_val = self.result_widget.sb_displacement_factor_max.value()
        res = round(value / max_val * 100)
        return res

    def convert_from_slider_value(self, value: float):
        if value == 0.0:
            return 0.0

        max_val = self.result_widget.sb_displacement_factor_max.value()
        return value / 100 * max_val

    def hsb_disp_factor_changed(self, value):
        inputbox = self.result_widget.sb_displacement_factor

        scaled_val = self.convert_from_slider_value(value)
        FreeCAD.FEM_dialog["disp_factor"] = scaled_val
        self.update_displacement()

        # Prevent recursive calls
        inputbox.blockSignals(True)
        inputbox.setValue(scaled_val)
        inputbox.blockSignals(False)

    def sb_disp_factor_max_changed(self, value):
        FreeCAD.FEM_dialog["disp_factor_max"] = value
        inputbox = self.result_widget.sb_displacement_factor
        inputbox.setMaximum(value)

        self.result_widget.hsb_displacement_factor.setValue(
            self.convert_to_slider_value(inputbox.value())
        )

    def sb_disp_factor_changed(self, value):
        FreeCAD.FEM_dialog["disp_factor"] = value
        self.update_displacement()

        slider = self.result_widget.hsb_displacement_factor

        # Prevent recursive calls
        slider.blockSignals(True)
        scaled_val = self.convert_to_slider_value(value)
        slider.setValue(scaled_val)
        slider.blockSignals(False)

    def set_focus_sb_disp(self):
        self.result_widget.sb_displacement_factor.selectAll()
        self.result_widget.sb_displacement_factor.setFocus()

    def disable_empty_result_buttons(self):
        """disable radio buttons if result does not exists in result object"""
        """assignments
        DisplacementLengths --> rb_abs_displacement
        DisplacementVectors --> rb_x_displacement, rb_y_displacement, rb_z_displacement
        Temperature         --> rb_temperature
        vonMises            --> rb_vm_stress
        PrincipalMax        --> rb_maxprin
        PrincipalMin        --> rb_minprin
        MaxShear            --> rb_max_shear_stress
        MassFlowRate        --> rb_massflowrate
        NetworkPressure     --> rb_networkpressure
        Peeq                --> rb_peeq"""
        if len(self.result_obj.DisplacementLengths) == 0:
            self.result_widget.rb_abs_displacement.setEnabled(0)
        if len(self.result_obj.DisplacementVectors) == 0:
            self.result_widget.rb_x_displacement.setEnabled(0)
            self.result_widget.rb_y_displacement.setEnabled(0)
            self.result_widget.rb_z_displacement.setEnabled(0)
        if len(self.result_obj.Temperature) == 0:
            self.result_widget.rb_temperature.setEnabled(0)
        if len(self.result_obj.vonMises) == 0:
            self.result_widget.rb_vm_stress.setEnabled(0)
        if len(self.result_obj.PrincipalMax) == 0:
            self.result_widget.rb_maxprin.setEnabled(0)
        if len(self.result_obj.PrincipalMin) == 0:
            self.result_widget.rb_minprin.setEnabled(0)
        if len(self.result_obj.MaxShear) == 0:
            self.result_widget.rb_max_shear_stress.setEnabled(0)
        if len(self.result_obj.MassFlowRate) == 0:
            self.result_widget.rb_massflowrate.setEnabled(0)
        if len(self.result_obj.NetworkPressure) == 0:
            self.result_widget.rb_networkpressure.setEnabled(0)
        if len(self.result_obj.Peeq) == 0:
            self.result_widget.rb_peeq.setEnabled(0)

    def update(self):
        self.reset_result_mesh()
        self.suitable_results = False
        self.disable_empty_result_buttons()
        if self.mesh_obj.FemMesh.NodeCount == 0:
            the_error_messagetext = (
                "FEM: there are no nodes in result mesh, there will be nothing to show."
            )
            error_message = translate("FEM", the_error_messagetext) + "\n"
            FreeCAD.Console.PrintError(error_message)
            QtGui.QMessageBox.critical(
                None, translate("FEM", "Result mesh is empty"), error_message
            )
        elif self.mesh_obj.FemMesh.NodeCount == len(self.result_obj.NodeNumbers):
            self.suitable_results = True
            hide_parts_constraints()
        else:
            if not self.mesh_obj.FemMesh.VolumeCount:
                the_error_messagetext = (
                    "FEM: Graphical bending stress output "
                    "for beam or shell FEM Meshes not yet supported."
                )
                error_message = translate("FEM", the_error_messagetext) + "\n"
                FreeCAD.Console.PrintError(error_message)
                QtGui.QMessageBox.critical(
                    None, translate("FEM", "No result object"), error_message
                )
            else:
                the_error_messagetext = (
                    "FEM: Result node numbers are not equal to FEM Mesh NodeCount."
                )
                error_message = translate("FEM", the_error_messagetext) + "\n"
                FreeCAD.Console.PrintError(error_message)
                QtGui.QMessageBox.critical(
                    None, translate("FEM", "No result object"), error_message
                )

    def reset_mesh_color(self):
        self.mesh_obj.ViewObject.NodeColor = {}
        self.mesh_obj.ViewObject.ElementColor = {}
        self.mesh_obj.ViewObject.resetNodeColor()

    def reset_result_mesh(self):
        self.mesh_obj.ViewObject.resetNodeDisplacement()
        self.reset_mesh_color()

    def reject(self):
        self.reset_result_mesh()
        plt.close()
        # if the tasks panel is called from Command obj is not in edit mode
        # thus reset edit does not close the dialog, maybe don't call but set in edit instead
        FreeCADGui.Control.closeDialog()
        FreeCADGui.ActiveDocument.resetEdit()
        if len(self.animateText) > 0:
            for a in self.animateText:
                a.hide()
            self.animateText = []
        self.startAnimate = False
        FreeCAD.FEM_dialog["animate"][0] = self.result_widget.steps.value()
        FreeCAD.FEM_dialog["animate"][1] = self.result_widget.loops.value()
        FreeCAD.FEM_dialog["animate"][2] = self.result_widget.framerate.value()

    # animation start
    def animate_displacement(self):
        if "result_obj" in FreeCAD.FEM_dialog:
            if FreeCAD.FEM_dialog["result_obj"] != self.result_obj:
                self.update_displacement()
        self.result_widget.cb_show_displacement.setChecked(True)
        FreeCAD.FEM_dialog["result_obj"] = self.result_obj
        if self.suitable_results:
            self.mesh_obj.ViewObject.setNodeDisplacementByVectors(
                self.result_obj.NodeNumbers, self.result_obj.DisplacementVectors
            )
        self.result_widget.startButton.setText("Stop Animation")
        frame_rate = 10
        self.hsb_displacement_factor = self.result_widget.sb_displacement_factor.value()
        frame_rate = self.result_widget.framerate.value()
        steps_per_cycle = int(self.result_widget.steps.value())
        number_cycles = int(self.result_widget.loops.value())

        inc = math.pi / steps_per_cycle * 2.0
        self.set_label(self.result_obj.Label, self.results_name)

        done = False
        for lo in range(0, number_cycles):
            for st in range(0, steps_per_cycle):
                self.mesh_obj.ViewObject.applyDisplacement(
                    math.sin(st * inc) * self.hsb_displacement_factor
                )
                FreeCADGui.updateGui()
                if not self.startAnimate:
                    done = True
                    break
                time.sleep(1.0 / frame_rate)  # modify the time here
            if done:
                break
        try:
            self.result_widget.startButton.setText("Start Animation")
        except:
            pass
        QtGui.QApplication.restoreOverrideCursor()
        self.startAnimate = False

    def value_changed(self, dummy, value, myType):
        # the only actions are:
        if myType == "startButton":
            if not self.startAnimate:
                self.startAnimate = True
                self.animate_displacement()
            else:
                self.startAnimate = False
        else:
            pass
        try:
            self.hsb_displacement_factor = self.result_widget.sb_displacement_factor.value()

        except:
            pass
        return

    def set_label(self, result_name, mesh_data):
        if len(self.animateText) == 0:
            self.animateText.append(CreateLabels.createLabel((-0.98, 0.90, 0), result_name))
            self.animateText.append(CreateLabels.createLabel((-0.98, 0.70, 0), mesh_data))
        else:
            self.animateText[1].set_text(mesh_data)
        pass


# animation end


# helper
def hide_parts_constraints():
    from FemGui import getActiveAnalysis

    fem_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/General")
    hide_constraints = fem_prefs.GetBool("HideConstraint", False)
    if hide_constraints:
        for o in FreeCAD.ActiveDocument.Objects:
            if o.isDerivedFrom("Fem::FemAnalysis"):
                for acnstrmesh in getActiveAnalysis().Group:
                    if "Constraint" in acnstrmesh.TypeId:
                        acnstrmesh.ViewObject.Visibility = False
                break


def get_displacement_scale_factor(res_obj):
    node_items = res_obj.Mesh.FemMesh.Nodes.items()
    displacements = res_obj.DisplacementVectors
    # use standard scale if there are no displacements in result object
    if len(displacements) == 0:
        return 1
    x_max, y_max, z_max = map(max, zip(*displacements))
    positions = []  # list of node vectors
    for k, v in node_items:
        positions.append(v)
    p_x_max, p_y_max, p_z_max = map(max, zip(*positions))
    p_x_min, p_y_min, p_z_min = map(min, zip(*positions))
    x_span = abs(p_x_max - p_x_min)
    y_span = abs(p_y_max - p_y_min)
    z_span = abs(p_z_max - p_z_min)
    span = max(x_span, y_span, z_span)
    max_disp = max(x_max, y_max, z_max)
    if max_disp == 0.0:
        return 0.0  # avoid float division by zero
    # FIXME - add max_allowed_disp to Preferences
    max_allowed_disp = 0.01 * span
    scale = max_allowed_disp / max_disp
    return scale



# SPDX-License-Identifier: LGPL-2.1-or-later

from typing import Any

from Base.Metadata import export

from App.DocumentObject import DocumentObject

@export(
    Include="Mod/CAM/App/FeaturePathCompound.h",
    TwinPointer="FeatureCompound",
    Namespace="Path",
)
class FeaturePathCompound(DocumentObject):
    """
    This class handles Path Compound features

    Author: Yorik van Havre (yorik@uncreated.net)
    License: LGPL-2.1-or-later
    """

    def addObject(self) -> Any:
        """Add an object to the group"""
        ...

    def removeObject(self) -> Any:
        """Remove an object from the group"""
        ...



#! python
# SPDX-License-Identifier: LGPL-2.1-or-later

# (c) 2018 Werner Mayer LGPL
#

import sys, getopt

# import os # The code that needs this is commented out
import shutil


def main():
    outputfile = ""
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:", ["outputfile="])
    except getopt.GetoptError:
        pass

    for o, a in opts:
        if o in ("-o", "--outputfile"):
            outputfile = a

    # if os.path.exists(outputfile):
    #    do_not_create = True
    #    ts = os.path.getmtime(outputfile)
    #    for f in args:
    #        if os.path.getmtime(f) > ts:
    #            do_not_create = False
    #            break
    #
    #    if do_not_create:
    #        print ("Up-to-date file {0}".format(outputfile))
    #        return

    with open(outputfile, "wb") as wfd:
        for f in args:
            with open(f, "rb") as fd:
                shutil.copyfileobj(fd, wfd, 1024 * 1024 * 10)
        print("Created file {0}".format(outputfile))


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import os
import tempfile
import unittest

from Addon import Addon, INTERNAL_WORKBENCHES
from addonmanager_macro import Macro


class TestAddon(unittest.TestCase):

    def setUp(self):
        self.test_dir = os.path.join(os.path.dirname(__file__), "..", "data")

    def test_display_name(self):
        # Case 1: No display name set elsewhere: name == display_name
        addon = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        self.assertEqual(addon.name, "FreeCAD")
        self.assertEqual(addon.display_name, "FreeCAD")

        # Case 2: Package.xml metadata file sets a display name:
        addon.load_metadata_file(os.path.join(self.test_dir, "good_package.xml"))
        self.assertEqual(addon.name, "FreeCAD")
        self.assertEqual(addon.display_name, "Test Workbench")

    def test_git_url_cleanup(self):
        base_url = "https://github.com/FreeCAD/FreeCAD"
        test_urls = [f"  {base_url}  ", f"{base_url}.git", f"  {base_url}.git  "]
        for url in test_urls:
            addon = Addon("FreeCAD", url, Addon.Status.NOT_INSTALLED, "master")
            self.assertEqual(addon.url, base_url)

    def test_tag_extraction(self):
        addon = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon.load_metadata_file(os.path.join(self.test_dir, "good_package.xml"))

        tags = addon.tags
        self.assertEqual(len(tags), 5)
        expected_tags = set()
        expected_tags.add("Tag0")
        expected_tags.add("Tag1")
        expected_tags.add("TagA")
        expected_tags.add("TagB")
        expected_tags.add("TagC")
        self.assertEqual(expected_tags, tags)

    def test_contains_functions(self):
        # Test package.xml combinations:

        # Workbenches
        addon_with_workbench = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon_with_workbench.load_metadata_file(os.path.join(self.test_dir, "workbench_only.xml"))
        self.assertTrue(addon_with_workbench.contains_workbench())
        self.assertFalse(addon_with_workbench.contains_macro())
        self.assertFalse(addon_with_workbench.contains_preference_pack())
        self.assertFalse(addon_with_workbench.contains_bundle())
        self.assertFalse(addon_with_workbench.contains_other())

        # Macros
        addon_with_macro = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon_with_macro.load_metadata_file(os.path.join(self.test_dir, "macro_only.xml"))
        self.assertFalse(addon_with_macro.contains_workbench())
        self.assertTrue(addon_with_macro.contains_macro())
        self.assertFalse(addon_with_macro.contains_preference_pack())
        self.assertFalse(addon_with_workbench.contains_bundle())
        self.assertFalse(addon_with_workbench.contains_other())

        # Preference Packs
        addon_with_prefpack = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon_with_prefpack.load_metadata_file(os.path.join(self.test_dir, "prefpack_only.xml"))
        self.assertFalse(addon_with_prefpack.contains_workbench())
        self.assertFalse(addon_with_prefpack.contains_macro())
        self.assertTrue(addon_with_prefpack.contains_preference_pack())
        self.assertFalse(addon_with_workbench.contains_bundle())
        self.assertFalse(addon_with_workbench.contains_other())

        # Combination
        addon_with_all = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon_with_all.load_metadata_file(os.path.join(self.test_dir, "combination.xml"))
        self.assertTrue(addon_with_all.contains_workbench())
        self.assertTrue(addon_with_all.contains_macro())
        self.assertTrue(addon_with_all.contains_preference_pack())
        self.assertTrue(addon_with_all.contains_bundle())
        self.assertTrue(addon_with_all.contains_other())

        # Now do the simple, explicitly-set cases
        addon_wb = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon_wb.repo_type = Addon.Kind.WORKBENCH
        self.assertTrue(addon_wb.contains_workbench())
        self.assertFalse(addon_wb.contains_macro())
        self.assertFalse(addon_wb.contains_preference_pack())

        addon_m = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon_m.repo_type = Addon.Kind.MACRO
        self.assertFalse(addon_m.contains_workbench())
        self.assertTrue(addon_m.contains_macro())
        self.assertFalse(addon_m.contains_preference_pack())

        # There is no equivalent for preference packs, they are always accompanied by a
        # metadata file

    def test_create_from_macro(self):
        macro_file = os.path.join(self.test_dir, "DoNothing.FCMacro")
        macro = Macro("DoNothing")
        macro.fill_details_from_file(macro_file)
        addon = Addon.from_macro(macro)

        self.assertEqual(addon.repo_type, Addon.Kind.MACRO)
        self.assertEqual(addon.name, "DoNothing")
        self.assertEqual(
            addon.macro.comment,
            "Do absolutely nothing. For Addon Manager integration tests.",
        )
        self.assertEqual(addon.url, "https://github.com/FreeCAD/FreeCAD")
        self.assertEqual(addon.macro.version, "1.0")
        self.assertEqual(len(addon.macro.other_files), 3)
        self.assertEqual(addon.macro.author, "Chris Hennes")
        self.assertEqual(addon.macro.date, "2022-02-28")
        self.assertEqual(addon.macro.icon, "not_real.png")
        self.assertNotEqual(addon.macro.xpm, "")

    def test_dependency_resolution(self):
        addonA = Addon(
            "AddonA",
            "https://github.com/FreeCAD/FakeAddonA",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addonB = Addon(
            "AddonB",
            "https://github.com/FreeCAD/FakeAddonB",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addonC = Addon(
            "AddonC",
            "https://github.com/FreeCAD/FakeAddonC",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addonD = Addon(
            "AddonD",
            "https://github.com/FreeCAD/FakeAddonD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )

        addonA.requires.add("AddonB")
        addonB.requires.add("AddonC")
        addonB.requires.add("AddonD")
        addonD.requires.add("CAM")

        all_addons = {
            addonA.name: addonA,
            addonB.name: addonB,
            addonC.name: addonC,
            addonD.name: addonD,
        }

        deps = Addon.Dependencies()
        addonA.walk_dependency_tree(all_addons, deps)

        self.assertEqual(len(deps.required_external_addons), 3)
        addon_strings = [addon.name for addon in deps.required_external_addons]
        self.assertTrue(
            "AddonB" in addon_strings,
            "AddonB not in required dependencies, and it should be.",
        )
        self.assertTrue(
            "AddonC" in addon_strings,
            "AddonC not in required dependencies, and it should be.",
        )
        self.assertTrue(
            "AddonD" in addon_strings,
            "AddonD not in required dependencies, and it should be.",
        )
        self.assertTrue(
            "CAM" in deps.internal_workbenches,
            "CAM not in workbench dependencies, and it should be.",
        )

    def test_internal_workbench_list(self):
        addon = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon.load_metadata_file(os.path.join(self.test_dir, "depends_on_all_workbenches.xml"))
        deps = Addon.Dependencies()
        addon.walk_dependency_tree({}, deps)
        self.assertEqual(len(deps.internal_workbenches), len(INTERNAL_WORKBENCHES))

    def test_version_check(self):
        addon = Addon(
            "FreeCAD",
            "https://github.com/FreeCAD/FreeCAD",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        addon.load_metadata_file(os.path.join(self.test_dir, "test_version_detection.xml"))

        self.assertEqual(
            len(addon.tags),
            1,
            "Wrong number of tags found: version requirements should have restricted to only one",
        )
        self.assertFalse(
            "TagA" in addon.tags,
            "Found 'TagA' in tags, it should have been excluded by version requirement",
        )
        self.assertTrue(
            "TagB" in addon.tags,
            "Failed to find 'TagB' in tags, it should have been included",
        )
        self.assertFalse(
            "TagC" in addon.tags,
            "Found 'TagA' in tags, it should have been excluded by version requirement",
        )

    def test_try_find_wbname_in_files_empty_dir(self):
        with tempfile.TemporaryDirectory() as mod_dir:
            # Arrange
            test_addon = Addon("test")
            test_addon.mod_directory = mod_dir
            os.mkdir(os.path.join(mod_dir, test_addon.name))

            # Act
            wb_name = test_addon.try_find_workbench_name_in_files()

            # Assert
            self.assertEqual(wb_name, "")

    def test_try_find_wbname_in_files_non_python_ignored(self):
        with tempfile.TemporaryDirectory() as mod_dir:
            # Arrange
            test_addon = Addon("test")
            test_addon.mod_directory = mod_dir
            base_path = os.path.join(mod_dir, test_addon.name)
            os.mkdir(base_path)
            file_path = os.path.join(base_path, "test.txt")
            with open(file_path, "w", encoding="utf-8") as f:
                f.write("Gui.addWorkbench(TestWorkbench())")

            # Act
            wb_name = test_addon.try_find_workbench_name_in_files()

            # Assert
            self.assertEqual(wb_name, "")

    def test_try_find_wbname_in_files_simple(self):
        with tempfile.TemporaryDirectory() as mod_dir:
            # Arrange
            test_addon = Addon("test")
            test_addon.mod_directory = mod_dir
            base_path = os.path.join(mod_dir, test_addon.name)
            os.mkdir(base_path)
            file_path = os.path.join(base_path, "test.py")
            with open(file_path, "w", encoding="utf-8") as f:
                f.write("Gui.addWorkbench(TestWorkbench())")

            # Act
            wb_name = test_addon.try_find_workbench_name_in_files()

            # Assert
            self.assertEqual(wb_name, "TestWorkbench")

    def test_try_find_wbname_in_files_subdir(self):
        with tempfile.TemporaryDirectory() as mod_dir:
            # Arrange
            test_addon = Addon("test")
            test_addon.mod_directory = mod_dir
            base_path = os.path.join(mod_dir, test_addon.name)
            os.mkdir(base_path)
            subdir = os.path.join(base_path, "subdirectory")
            os.mkdir(subdir)
            file_path = os.path.join(subdir, "test.py")
            with open(file_path, "w", encoding="utf-8") as f:
                f.write("Gui.addWorkbench(TestWorkbench())")

            # Act
            wb_name = test_addon.try_find_workbench_name_in_files()

            # Assert
            self.assertEqual(wb_name, "TestWorkbench")

    def test_try_find_wbname_in_files_variable_used(self):
        with tempfile.TemporaryDirectory() as mod_dir:
            # Arrange
            test_addon = Addon("test")
            test_addon.mod_directory = mod_dir
            base_path = os.path.join(mod_dir, test_addon.name)
            os.mkdir(base_path)
            file_path = os.path.join(base_path, "test.py")
            with open(file_path, "w", encoding="utf-8") as f:
                f.write("Gui.addWorkbench(wb)")

            # Act
            wb_name = test_addon.try_find_workbench_name_in_files()

            # Assert
            self.assertEqual(wb_name, "")

    def test_try_find_wbname_in_files_variants(self):
        variants = [
            "Gui.addWorkbench(TestWorkbench())",
            "Gui.addWorkbench (TestWorkbench())",
            "Gui.addWorkbench( TestWorkbench() )",
            "Gui.addWorkbench(TestWorkbench( ))",
            "Gui.addWorkbench( TestWorkbench( ) )",
            "Gui.addWorkbench( TestWorkbench ( ) )",
            "Gui.addWorkbench ( TestWorkbench ( ) )",
        ]
        for variant in variants:
            with self.subTest(variant=variant):
                with tempfile.TemporaryDirectory() as mod_dir:
                    # Arrange
                    test_addon = Addon("test")
                    test_addon.mod_directory = mod_dir
                    base_path = os.path.join(mod_dir, test_addon.name)
                    os.mkdir(base_path)
                    file_path = os.path.join(base_path, "test.py")
                    with open(file_path, "w", encoding="utf-8") as f:
                        f.write(variant)

                    # Act
                    wb_name = test_addon.try_find_workbench_name_in_files()

                    # Assert
                    self.assertEqual(wb_name, "TestWorkbench")

    def test_addon_less_than(self):
        a = Addon("A")
        b = Addon("B")
        self.assertLess(a, b)

        aa = Addon("A", branch="A")
        ab = Addon("A", branch="B")
        self.assertLess(aa, ab)

        same = Addon("A", branch="A")
        self.assertGreaterEqual(same, aa)
        self.assertGreaterEqual(aa, same)

    def test_addon_less_than_or_equal_to(self):
        a = Addon("A")
        b = Addon("B")
        self.assertLessEqual(a, b)

        aa = Addon("A", branch="A")
        ab = Addon("A", branch="B")
        self.assertLessEqual(aa, ab)

        same = Addon("A", branch="A")
        self.assertLessEqual(same, aa)
        self.assertLessEqual(aa, same)



from Base.Metadata import export

from App.DocumentObjectGroup import DocumentObjectGroup

@export(Include="Mod/Assembly/App/SimulationGroup.h", Namespace="Assembly")
class SimulationGroup(DocumentObjectGroup):
    """
    This class is a group subclass for joints.

    Author: Ondsel (development@ondsel.com)
    License: LGPL-2.1-or-later
    """



from Base.Metadata import export
from Part.Conic2d import Conic2d
from typing import Final, overload

@export(
    Twin="Geom2dHyperbola",
    TwinPointer="Geom2dHyperbola",
    PythonName="Part.Geom2d.Hyperbola2d",
    FatherInclude="Mod/Part/App/Geom2d/Conic2dPy.h",
    Include="Mod/Part/App/Geometry2d.h",
    Constructor=True,
)
class Hyperbola2d(Conic2d):
    """
    Describes a hyperbola in 2D space
    To create a hyperbola there are several ways:
    Part.Geom2d.Hyperbola2d()
        Creates a hyperbola with major radius 2 and minor radius 1 with the
        center in (0,0)

    Part.Geom2d.Hyperbola2d(Hyperbola)
        Create a copy of the given hyperbola

    Part.Geom2d.Hyperbola2d(S1,S2,Center)
        Creates a hyperbola centered on the point Center, S1 and S2,
        its major axis is defined by Center and S1,
        its major radius is the distance between Center and S1, and
        its minor radius is the distance between S2 and the major axis.

    Part.Geom2d.Hyperbola2d(Center,MajorRadius,MinorRadius)
        Creates a hyperbola with major and minor radii MajorRadius and
        MinorRadius and located at Center

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    MajorRadius: float = ...
    """The major radius of the hyperbola."""

    MinorRadius: float = ...
    """The minor radius of the hyperbola."""

    Focal: Final[float] = ...
    """The focal distance of the hyperbola."""

    Focus1: Final[object] = ...
    """
    The first focus is on the positive side of the major axis of the hyperbola;
    the second focus is on the negative side.
    """

    Focus2: Final[object] = ...
    """
    The first focus is on the positive side of the major axis of the hyperbola;
    the second focus is on the negative side.
    """

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, Hyperbola: "Hyperbola2d") -> None: ...
    @overload
    def __init__(self, S1: object, S2: object, Center: object) -> None: ...
    @overload
    def __init__(self, Center: object, MajorRadius: float, MinorRadius: float) -> None: ...



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
from CAMTests.PathTestUtils import PathTestBase


class TestPathCore(PathTestBase):
    def test00(self):
        """Test Path command core functionality"""
        # create empty command
        c = Path.Command()
        self.assertIsInstance(c, Path.Command)

        # change name
        c.Name = "G1"
        self.assertEqual(c.Name, "G1")

        # Assign Parameters
        c.Parameters = {"X": 1, "Y": 0}
        self.assertEqual(c.Parameters, {"Y": 0.0, "X": 1.0})

        # change parameters
        c.Parameters = {"X": 1, "Y": 0.5}
        self.assertEqual(c.Parameters, {"Y": 0.5, "X": 1})

        # output gcode
        self.assertEqual(c.toGCode(), "G1 X1.000000 Y0.500000")

        # create and assign name in one
        c2 = Path.Command("G2")
        self.assertEqual(c2.Name, "G2")

        # Create Path and parameters in one
        c3 = Path.Command("G1", {"X": 34, "Y": 1.2})
        self.assertEqual(str(c3), "Command G1 [ X:34 Y:1.2 ]")
        c4 = Path.Command("G1X4Y5")
        self.assertEqual(str(c4), "Command G1 [ X:4 Y:5 ]")

        # use placement
        self.assertEqual(str(c3.Placement), "Placement [Pos=(34,1.2,0), Yaw-Pitch-Roll=(0,0,0)]")
        self.assertEqual(c3.toGCode(), "G1 X34.000000 Y1.200000")
        p1 = FreeCAD.Placement()
        p1.Base = FreeCAD.Vector(3, 2, 1)
        self.assertEqual(str(p1), "Placement [Pos=(3,2,1), Yaw-Pitch-Roll=(0,0,0)]")
        c5 = Path.Command("g1", p1)
        self.assertEqual(str(c5), "Command G1 [ X:3 Y:2 Z:1 ]")
        p2 = FreeCAD.Placement()
        p2.Base = FreeCAD.Vector(5, 0, 0)

        # overwrite placement
        c5.Placement = p2
        self.assertEqual(str(c5), "Command G1 [ X:5 ]")
        self.assertEqual(c5.x, 5.0)

        # overwrite individual parameters
        c5.x = 10
        self.assertEqual(c5.x, 10.0)
        c5.y = 2
        self.assertEqual(str(c5), "Command G1 [ X:10 Y:2 ]")

        # set from gcode
        c3.setFromGCode("G1X1Y0")
        self.assertEqual(str(c3), "Command G1 [ X:1 Y:0 ]")

    def test10(self):
        """Test Path Object core functionality"""

        c1 = Path.Command("g1", {"x": 1, "y": 0})
        c2 = Path.Command("g1", {"x": 0, "y": 2})
        p = Path.Path([c1, c2])
        self.assertAlmostEqual(str(p), "Path [ size:2 length:3.2361 ]", places=4)

        self.assertEqual(str(p.Commands), "[Command G1 [ X:1 Y:0 ], Command G1 [ X:0 Y:2 ]]")
        self.assertAlmostEqual(p.Length, 3.2361, places=4)
        p.addCommands(c1)
        self.assertEqual(
            p.toGCode(),
            "G1 X1.000000 Y0.000000\nG1 X0.000000 Y2.000000\nG1 X1.000000 Y0.000000\n",
        )

        lines = """
G0X-0.5905Y-0.3937S3000M03
G0Z0.125
G1Z-0.004F3
G1X0.9842Y-0.3937F14.17
G1X0.9842Y0.433
G1X-0.5905Y0.433
G1X-0.5905Y-0.3937
G0Z0.5
"""

        output = """G0 S3000.000000 X-0.590500 Y-0.393700
M03
G0 Z0.125000
G1 F3.000000 Z-0.004000
G1 F14.170000 X0.984200 Y-0.393700
G1 X0.984200 Y0.433000
G1 X-0.590500 Y0.433000
G1 X-0.590500 Y-0.393700
G0 Z0.500000
"""

        # create a path directly form a piece of gcode.
        p = Path.Path()
        p.setFromGCode(lines)
        self.assertEqual(p.toGCode(), output)

    def test50(self):
        """Test Path.Length calculation"""
        commands = []
        commands.append(Path.Command("G1", {"X": 1}))
        commands.append(Path.Command("G1", {"Y": 1}))
        path = Path.Path(commands)

        self.assertEqual(path.Length, 2)



# SPDX-License-Identifier: BSD-3-Clause

def process_nested_parent_str(attr_str):
    """
    The first letter should be a parenthesis
    input string: "(1,4,(5,6),7)"
    output: tuple (1,4,(4,6),7)
    """
    params = []
    agg_scope_level = 0
    current_param = ""
    for i, ch in enumerate(attr_str):
        if ch == ",":
            params.append(current_param)
            current_param = ""
        elif ch == "(":
            agg_scope_level += 1
        elif ch == ")":
            agg_scope_level = 0
        elif agg_scope_level == 0:
            current_param += ch
    return params


def process_nested_parent_str2(attr_str, idx=0):
    """
    The first letter should be a parenthesis
    input string: "(1,4,(5,6),7)"
    output: ['1','4',['5','6'],'7']
    """
    # print 'Entering function with string %s'%(attr_str)
    params = []
    current_param = ""
    k = 0
    while k < len(attr_str):
        # print 'k in this function:%i'%k
        ch = attr_str[k]
        k += 1
        if ch == ",":
            # print "Add param:",current_param
            params.append(current_param)
            current_param = ""
        elif ch == "(":
            nv = attr_str[k:]
            # print "Up one level parenthesis:%s"%(nv)
            current_param, progress = process_nested_parent_str2(nv)
            # print "Adding the list returned from nested",current_param
            params.append(current_param)
            current_param = ""
            k += progress + 1
        elif ch == ")":
            # print "Down one level parenthesis: %i characters parsed"%k
            params.append(current_param)
            # print "Current params:",params#k -= acc-2
            return params, k
        else:
            current_param += ch
        # print "Ch:",ch
        # print "k:",k

        # raw_input("")
        # idx += 1

    params.append(current_param)
    return params, k


# print process_nested_parent_str2('1,2,3,4,5,6')
# idx=0
# print process_nested_parent_str2("'A','B','C'")
print(process_nested_parent_str2("'A'")[0])
print(process_nested_parent_str2("30.0,0.0,5.0")[0])
print(process_nested_parent_str2("(Thomas)")[0])
print(process_nested_parent_str2("Thomas, Paviot, ouais")[0])
print(process_nested_parent_str2("1,2,(3,4,5),6,7,8")[0])
print(process_nested_parent_str2("(#9149,#9166),#9142,.T.")[0])



from Base.Metadata import export, constmethod, no_args
from Base.Quantity import Quantity
from Base.Vector import Vector
from Base.Axis import Axis
from Part.App.Part2DObject import Part2DObject
from Part.App.Geometry import Geometry
from Sketcher.App.Constraint import Constraint
from typing import List, Tuple, Union, Final, overload

@export(
    Include="Mod/Sketcher/App/SketchObject.h",
    FatherInclude="Mod/Part/App/Part2DObjectPy.h",
)
class SketchObject(Part2DObject):
    """
    Represents a sketch object

    Author: Juergen Riegel
    Licence: LGPL
    """

    MissingPointOnPointConstraints: List = ...
    """Returns a list of (First FirstPos Second SecondPos Type) tuples with all the detected endpoint constraints."""

    MissingVerticalHorizontalConstraints: List = ...
    """Returns a list of (First FirstPos Second SecondPos Type) tuples with all the detected vertical/horizontal constraints."""

    MissingLineEqualityConstraints: List = ...
    """Returns a list of (First FirstPos Second SecondPos) tuples with all the detected line segment equality constraints."""

    MissingRadiusConstraints: List = ...
    """Returns a list of (First FirstPos Second SecondPos) tuples with all the detected radius constraints."""

    OpenVertices: Final[List] = ...
    """Returns a list of vertices positions."""

    ConstraintCount: Final[int] = ...
    """Number of Constraints in this sketch"""

    GeometryCount: Final[int] = ...
    """Number of geometric objects in this sketch"""

    AxisCount: Final[int] = ...
    """Return the number of construction lines in the sketch which can be used as axes"""

    GeometryFacadeList: List = ...
    """Return a list of GeometryFacade objects corresponding to the PropertyGeometryList"""

    DoF: Final[int] = ...
    """Return the DoFs of the current solved sketch"""

    ConflictingConstraints: Final[List] = ...
    """Return a list of integers indicating the constraints detected as conflicting"""

    RedundantConstraints: Final[List] = ...
    """Return a list of integers indicating the constraints detected as redundant"""

    PartiallyRedundantConstraints: Final[List] = ...
    """Return a list of integers indicating the constraints detected as partially redundant"""

    MalformedConstraints: Final[List] = ...
    """Return a list of integers indicating the constraints detected as malformed"""

    def solve(self) -> int:
        """
        Solve the sketch and update the geometry.

        solve()

          Returns:
              0 in case of success, otherwise the following codes in this order of
              priority:
              -4 if over-constrained,
              -3 if conflicting constraints,
              -5 if malformed constraints
              -1 if solver error,
              -2 if redundant constraints.
        """
        ...

    @overload
    def addGeometry(self, geo: Geometry, isConstruction: bool = False) -> int: ...
    @overload
    def addGeometry(self, geo: List[Geometry], isConstruction: bool = False) -> Tuple[int, ...]: ...
    def addGeometry(
        self, geo: Union[Geometry, List[Geometry]], isConstruction: bool = False
    ) -> Union[int, Tuple[int, ...]]:
        """
        Add geometric objects to the sketch.

        addGeometry(geo:Geometry, isConstruction=False) -> int
            Add a single geometric object to the sketch.

            Args:
                geo: The geometry to add. e.g. a Part.LineSegement
                isConstruction: Whether the added geometry is a "construction geometry".
                    Defaults to `False`, i.e. by omitting, a regular geometry is added.

            Returns:
                The zero-based index of the newly added geometry.

        addGeometry(geo:List(Geometry), isConstruction=False) -> Tuple(int)
            Add many geometric objects to the sketch.

            Args:
                geo: The geometry to add.
                isConstruction: see above.

            Returns:
                A tuple of zero-based indices of all newly added geometry.
        """
        ...

    def delGeometry(self, geoId: int, noSolve: bool) -> None:
        """
        Delete a geometric object from the sketch.

        delGeometry(geoId:int)

            Args:
                geoId: The zero-based index of the geometry to delete.
                    Any internal alignment geometry thereof will be deleted, too.
        """
        ...

    def delGeometries(self, geoIds: List[int], noSolve: bool) -> None:
        """
        Delete a list of geometric objects from the sketch.

        delGeometries(geoIds:List(int))

            Args:
                geoId: A list of zero-based indices of the geometry to delete.
                    Any internal alignment geometry thereof will be deleted, too.
        """
        ...

    def deleteAllGeometry(self, noSolve: bool) -> None:
        """
        Delete all the geometry objects from the sketch, except external geometry.

        deleteAllGeometry()
        """
        ...

    def detectDegeneratedGeometries(self, tolerance: float) -> int:
        """
        Detect degenerated geometries. A curve geometry is considered degenerated
        if the parameter range is less than the tolerance.

        detectDegeneratedGeometries(tolerance:float)

            Args:
                tolerance: The tolerance to check the parameter range of a curve.

            Returns:
                The number of degenerated geometries.
        """
        ...

    def removeDegeneratedGeometries(self, tolerance: float) -> int:
        """
        Remove degenerated geometries. A curve geometry is considered degenerated
        if the parameter range is less than the tolerance.

        removeDegeneratedGeometries(tolerance:float)

            Args:
                tolerance: The tolerance to check the parameter range of a curve.

            Returns:
                The number of degenerated geometries.
        """
        ...

    def deleteAllConstraints(self) -> None:
        """
        Delete all the constraints from the sketch.

        deleteAllConstraints()
        """
        ...

    def toggleConstruction(self, geoId: int) -> None:
        """
        Toggles a geometry between regular and construction.

        toggleConstruction(geoId:int)

            Args:
                geoId: The zero-based index of the geometry to toggle.
        """
        ...

    def setConstruction(self, geoId: int, state: bool) -> None:
        """
        Set construction mode of a geometry.

        setConstruction(geoId:int, state:bool)

            Args:
                geoId: The zero-based index of the geometry to configure.
                state: `True` configures the geometry to "construction geometry",
                    `False` configures it to regular geometry.
        """
        ...

    def getConstruction(self, geoId: int) -> bool:
        """
        Determine whether the given geometry is a "construction geometry".

        getConstruction(geoId:int)

            Args:
                geoId: The zero-based index of the geometry to query.

            Returns:
                `True` if the geometry is "construction geometry" and
                `False` if it s a regular geometry.
        """
        ...

    @overload
    def addConstraint(self, constraint: Constraint) -> int: ...
    @overload
    def addConstraint(self, constraints: List[Constraint]) -> Tuple[int, ...]: ...
    def addConstraint(
        self, constraint: Union[Constraint, List[Constraint]]
    ) -> Union[int, Tuple[int, ...]]:
        """
        Add constraints to the sketch.

        addConstraint(constraint:Constraint) -> int
            Add a single constraint to the sketch and solves it.

            Returns:
                The zero-based index of the newly added constraint.

        addConstraint(constraints:List(Constraint)) -> Tuple(int)
            Add many constraints to the sketch without solving.

            Returns:
                A tuple of zero-based indices of all newly added constraints.
        """
        ...

    def delConstraint(self, constraintIndex: int, noSolve: bool) -> None:
        """
        Delete a constraint from the sketch.

        delConstraint(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to delete.
        """
        ...

    def delConstraints(
        self, constraintIndices: List[int], updateGeometry: bool, noSolve: bool
    ) -> None:
        """
        Delete multiple constraints from a sketch

        delConstraints(constraintIndices: List[int], updateGeometry: bool)

            Args:
                constraintIndices: The zero-based indices of the constraints to delete
                updateGeometry: Whether to update the geometry after solve
        """
        ...

    def renameConstraint(self, constraintIndex: int, name: str) -> None:
        """
        Rename a constraint in the sketch.

        renameConstraint(constraintIndex:int, name:str)

            Args:
                constraintIndex: The zero-based index of the constraint to rename.
                name: The new name for the constraint.
                    An empty string makes the constraint "unnamed" again.
        """
        ...

    @constmethod
    def getIndexByName(self, name: str) -> int:
        """
        Get the index of a constraint by name.

        getIndexByName(name:str)

            Args:
                name: The name for the constraint to look up.
                    If there is no such constraint an exception is raised.
        """
        ...

    def carbonCopy(self, objName: str, asConstruction: bool = True) -> None:
        """
        Copy another sketch's geometry and constraints into this sketch.

        carbonCopy(objName:str, asConstruction=True)

            Args:
                ObjName: The name of the sketch object to copy from.
                asConstruction: Whether to copy the geometry as "construction geometry".
        """
        ...

    def addExternal(
        self, objName: str, subName: str, defining: bool = False, intersection: bool = False
    ) -> None:
        """
        Add a link to an external geometry.

        addExternal(objName:str, subName:str, defining:bool=False, intersection:bool=False)

            Args:
                objName: The name of the document object to reference.
                subName: The name of the sub-element of the object's shape to link as
                    "external geometry".
                defining: Should the external edges be defining or construction?
                intersection: Should the external edges be projections or intersections?
        """
        ...

    def delExternal(self, extGeoId: int) -> None:
        """
        Delete an external geometry link from the sketch.

        delExternal(extGeoId:int)

            Args:
                extGeoId: The zero-based index of the external geometry to remove.
        """
        ...

    @overload
    def delConstraintOnPoint(self, vertexId: int) -> None: ...
    @overload
    def delConstraintOnPoint(self, geoId: int, pointPos: int) -> None: ...
    def delConstraintOnPoint(self, *args: int) -> None:
        """
        Delete coincident constraints associated with a sketch point.

        delConstraintOnPoint(vertexId:int)

            Args:
                vertexId: A zero-based index of the shape's vertices.

        delConstraintOnPoint(geoId:int, pointPos:int)

            Args:
                geoId: The zero-based index of the geometry that contains the point.
                pointPos: Enum denoting which point on the geometry is meant:
                    1: the start of a line or bounded curve.
                    2: the end of a line or bounded curve.
                    3: the center of a circle or ellipse.
        """
        ...

    @no_args
    def delConstraintsToExternal(self) -> None:
        """
        Deletes all constraints referencing an external geometry.
        """
        ...

    def setDatum(self, constraint: Union[int, str], value: Union[float, Quantity]) -> None:
        """
        Set the value of a datum constraint (e.g. Distance or Angle)

        setDatum(constraint, value)

            Args:
                constraint (int or str): The index or name of the constraint to set.
                value (float or Quantity): The value to set for the constraint. When
                    using floats, values for linear dimensions are interpreted as
                    millimeter, angular ones as radians.
        """
        ...

    @constmethod
    def getDatum(self, constraint: Union[int, str]) -> Quantity:
        """
        Get the value of a datum constraint (e.g. Distance or Angle)

        getDatum(constraint) -> Quantity

            Args:
                constraint (int or str): The index or name of the constraint to query.

            Returns:
                The value of the constraint.
        """
        ...

    def setDriving(self, constraintIndex: int, state: bool) -> None:
        """
        Set the Driving status of a datum constraint.

        setDriving(constraintIndex:int, state:bool)

            Args:
                constraintIndex: The zero-based index of the constraint to configure.
                state: `True` sets the constraint to driving,
                    `False` configures it as non-driving, i.e. reference.
        """
        ...

    def setDatumsDriving(self, state: bool) -> None:
        """
        Set the Driving status of all datum constraints.

        setDatumsDriving(state:bool)

            Args:
                state: `True` set all datum constraints to driving,
                    `False` configures them as non-driving, i.e. reference.
        """
        ...

    def moveDatumsToEnd(self) -> None:
        """
        Moves all datum constraints to the end of the constraint list.

        moveDatumsToEnd()

            Warning: This method reorders the constraint indices. Previously held
                numeric references to constraints may reference different constraints
                after this operation.
        """
        ...

    @constmethod
    def getDriving(self, constraintIndex: int) -> bool:
        """
        Get the Driving status of a datum constraint.

        getDriving(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to query.

            Returns:
                `True` if the constraint is driving,
                `False` if it is non-driving, i.e. reference.
        """
        ...

    def toggleDriving(self, constraintIndex: int) -> None:
        """
        Toggle the Driving status of a datum constraint.

        toggleDriving(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to toggle.
        """
        ...

    def setVirtualSpace(self) -> None:
        """
        Set the VirtualSpace status of a constraint
        """
        ...

    def setVisibility(self) -> None:
        """
        Set the visibility of a constraint
        """
        ...

    def getVirtualSpace(self) -> bool:
        """
        Get the VirtualSpace status of a constraint
        """
        ...

    def toggleVirtualSpace(self) -> None:
        """
        Toggle the VirtualSpace status of a constraint
        """
        ...

    def setActive(self, constraintIndex: int, state: bool) -> None:
        """
        Activates or deactivates a constraint (enforce it or not).

        setActive(constraintIndex:int, state:bool)

            Args:
                constraintIndex: The zero-based index of the constraint to configure.
                state: `True` sets the constraint to active i.e. enforced,
                    `False` configures it as inactive, i.e. not enforced.
        """
        ...

    @constmethod
    def getActive(self, constraintIndex: int) -> bool:
        """
        Get whether a constraint is active, i.e. enforced, or not.

        getActive(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to query.

            Returns:
                `True` if the constraint is active, i.e. enforced,
                `False` if it is inactive, i.e. not enforced.
        """
        ...

    def toggleActive(self, constraintIndex: int) -> None:
        """
        Toggle the constraint between active (enforced) and inactive.

        toggleActive(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to toggle.
        """
        ...

    @constmethod
    def getLabelPosition(self, constraintIndex: int) -> float:
        """
        Get label position of the constraint.

        getLabelPosition(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to query.

            Returns:
                float with the current value.
        """
        ...

    def setLabelPosition(self, constraintIndex: int, value: float) -> None:
        """
        Set label position of the constraint.

        setLabelPosition(constraintIndex:int, value:float)

            Args:
                constraintIndex: The zero-based index of the constraint to query.
                value: Value of the label position.
        """
        ...

    @constmethod
    def getLabelDistance(self, constraintIndex: int) -> float:
        """
        Get label distance of the constraint.

        getLabelDistance(constraintIndex:int)

            Args:
                constraintIndex: The zero-based index of the constraint to query.

            Returns:
                float with the current value.
        """
        ...

    def setLabelDistance(self, constraintIndex: int, value: float) -> None:
        """
        Set label distance of the constraint.

        setLabelDistance(constraintIndex:int, value:float)

            Args:
                constraintIndex: The zero-based index of the constraint to query.
                value: Value of the label position.
        """
        ...

    def moveGeometry(
        self, GeoIndex: int, PointPos: int, Vector: Vector, relative: bool = False
    ) -> None:
        """
        Move a given point (or curve) to another location.

        moveGeometry(GeoIndex,PointPos,Vector,[relative])

        It moves the specified point (or curve) to the given location by adding some
        temporary weak constraints and solving the sketch.
        This method is mostly used to allow the user to drag some portions of the sketch
        in real time by e.g. the mouse and it works only for underconstrained portions of
        the sketch.
        The argument 'relative', if present, states if the new location is given
        relatively to the current one.
        """
        ...

    def moveGeometries(
        self, Geos: List[Tuple[int, int]], Vector: Vector, relative: bool = False
    ) -> None:
        """
        Move given points and curves to another location.

        moveGeometries(Geos,Vector,[relative])

        It moves the specified points and curves to the given location by adding some
        temporary weak constraints and solving the sketch.
        This method is mostly used to allow the user to drag some portions of the sketch
        in real time by e.g. the mouse and it works only for underconstrained portions of
        the sketch.
        The argument 'relative', if present, states if the new location is given
        relatively to the current one. For group dragging this is enforced.
        Geos is a vector of pairs of geoId and posId.
        """
        ...

    @constmethod
    def getPoint(self, GeoIndex: int, PointPos: int) -> Vector:
        """
        Retrieve the vector of a point in the sketch.

        getPoint(GeoIndex,PointPos)
        """
        ...

    @constmethod
    def getGeoVertexIndex(self, index: int) -> Tuple[int, int]:
        """
        Retrieve the GeoId and PosId of a point in the sketch.

        (geoId, posId) = getGeoVertexIndex(index)
        """
        ...

    @constmethod
    def getAxis(self) -> Axis:
        """
        Return an axis based on the corresponding construction line
        """
        ...

    def fillet(self) -> None:
        """
        Create a fillet between two edges or at a point
        """
        ...

    def trim(self) -> None:
        """
        Trim a curve with a given id at a given reference point
        """
        ...

    def extend(self) -> None:
        """
        Extend a curve to new start and end positions
        """
        ...

    def split(self) -> None:
        """
        Split a curve with a given id at a given reference point
        """
        ...

    def join(self) -> None:
        """
        Join two curves at the given end points
        """
        ...

    def addSymmetric(self) -> None:
        """
        Add symmetric geometric objects to the sketch with respect to a reference point or line
        """
        ...

    def addCopy(self) -> None:
        """
        Add a copy of geometric objects to the sketch displaced by a vector3d
        """
        ...

    def addMove(self) -> None:
        """
        Move the geometric objects in the sketch displaced by a vector3d
        """
        ...

    def addRectangularArray(self) -> None:
        """
        Add an array of size cols by rows where each element is a copy of the selected geometric objects displaced by a vector3d in the cols direction and by a vector perpendicular to it in the rows direction
        """
        ...

    def removeAxesAlignment(self) -> None:
        """
        Modifies constraints so that the shape is not forced to be aligned with axes.
        """
        ...

    def ExposeInternalGeometry(self) -> None:
        """
        Deprecated -- use exposeInternalGeometry
        """
        ...

    def DeleteUnusedInternalGeometry(self) -> None:
        """
        Deprecated -- use deleteUnusedInternalGeometry
        """
        ...

    def exposeInternalGeometry(self) -> None:
        """
        Exposes all internal geometry of an object supporting internal geometry
        """
        ...

    def deleteUnusedInternalGeometry(self) -> None:
        """
        Deletes all unused (not further constrained) internal geometry
        """
        ...

    def convertToNURBS(self) -> None:
        """
        Approximates the given geometry with a B-spline
        """
        ...

    def increaseBSplineDegree(self) -> None:
        """
        Increases the given B-spline Degree by a number of degrees
        """
        ...

    def decreaseBSplineDegree(self) -> None:
        """
        Decreases the given B-spline Degree by a number of degrees by approximating this curve
        """
        ...

    def modifyBSplineKnotMultiplicity(self) -> None:
        """
        Increases or reduces the given BSpline knot multiplicity
        """
        ...

    def insertBSplineKnot(self) -> None:
        """
        Inserts a knot into the BSpline at the given param with given multiplicity. If the knot already exists, this increases the knot multiplicity by the given multiplicity.
        """
        ...

    def calculateAngleViaPoint(self, GeoId1: int, GeoId2: int, px: float, py: float) -> float:
        """
        calculateAngleViaPoint(GeoId1, GeoId2, px, py) - calculates angle between
        curves identified by GeoId1 and GeoId2 at point (x,y). The point must be
        on intersection of the curves, otherwise the result may be useless (except
        line-to-line, where (0,0) is OK). Returned value is in radians.
        """
        ...

    def isPointOnCurve(self, GeoIdCurve: int, x: float, y: float) -> bool:
        """
        isPointOnCurve(GeoIdCurve, float x, float y) -> bool - tests if the point (x,y)
        geometrically lies on a curve (e.g. ellipse). It treats lines as infinite,
        arcs as full circles/ellipses/etc.
        """
        ...

    def calculateConstraintError(self, index: int) -> float:
        """
        calculateConstraintError(index) - calculates the error function of the
        constraint identified by its index and returns the signed error value.
        The error value roughly corresponds to by how much the constraint is
        violated. If the constraint internally has more than one error function,
        the returned value is RMS of all errors (sign is lost in this case).
        """
        ...

    def changeConstraintsLocking(self, bLock: bool) -> None:
        """
        changeConstraintsLocking(bLock) - locks or unlocks all tangent and
        perpendicular constraints. (Constraint locking prevents it from
        flipping to another valid configuration, when e.g. external geometry
        is updated from outside.) The sketch solve is not triggered by the
        function, but the SketchObject is touched (a recompute will be
        necessary). The geometry should not be affected by the function.

        The bLock argument specifies, what to do. If true, all constraints
        are unlocked and locked again. If false, all tangent and perp.
        constraints are unlocked.
        """
        ...

    def getGeometryWithDependentParameters(self) -> List[Tuple[int, int]]:
        """
        getGeometryWithDependentParameters - returns a list of geoid posid pairs
        with all the geometry element edges and vertices which the solver regards
        as being dependent on other parameters.
        """
        ...

    def autoconstraint(self) -> None:
        """
        Automatic sketch constraining algorithm.
        """
        ...

    def detectMissingPointOnPointConstraints(self) -> None:
        """
        Detects missing Point On Point Constraints. The detect step just identifies possible missing constraints.
        The result may be retrieved or applied using the corresponding Get / Make methods.
        """
        ...

    def analyseMissingPointOnPointCoincident(self) -> None:
        """
        Analyses the already detected missing Point On Point Constraints to detect endpoint tangency/perpendicular.
        The result may be retrieved or applied using the corresponding Get / Make methods.
        """
        ...

    def detectMissingVerticalHorizontalConstraints(self) -> None:
        """
        Detects missing Horizontal/Vertical Constraints. The detect step just identifies possible missing constraints.
        The result may be retrieved or applied using the corresponding Get / Make methods.
        """
        ...

    def detectMissingEqualityConstraints(self) -> None:
        """
        Detects missing Equality Constraints. The detect step just identifies possible missing constraints.
        The result may be retrieved or applied using the corresponding Get / Make methods.
        """
        ...

    def makeMissingPointOnPointCoincident(self, arg: bool) -> None:
        """
        Applies the detected / set Point On Point coincident constraints. If the argument is True, then solving and redundant removal is done after each individual addition.
        """
        ...

    def makeMissingVerticalHorizontal(self, arg: bool) -> None:
        """
        Applies the detected / set Vertical/Horizontal constraints. If the argument is True, then solving and redundant removal is done after each individual addition.
        """
        ...

    def makeMissingEquality(self, arg: bool) -> None:
        """
        Applies the detected / set Equality constraints. If the argument is True, then solving and redundant removal is done after each individual addition.
        """
        ...

    @constmethod
    @no_args
    def evaluateConstraints(self) -> bool:
        """
        Check for constraints with invalid indexes. Returns True if invalid constraints are found, False otherwise.
        """
        ...

    @no_args
    def validateConstraints(self) -> None:
        """
        Removes constraints with invalid indexes.
        """
        ...

    def autoRemoveRedundants(self, arg: bool) -> None:
        """
        Removes constraints currently detected as redundant by the solver. If the argument is True, then the geometry is updated after solving.
        """
        ...

    def toPythonCommands(self) -> None:
        """
        Prints the commands that should be executed to recreate the Geometry and Constraints of the present sketch (excluding any External Geometry).
        """
        ...

    def setGeometryId():
        """
        Sets the GeometryId of the SketchGeometryExtension of the geometry with the provided GeoId
        """
        ...

    def setGeometryIds(GeoIdsToIds: List[Tuple[int, int]]):
        """
        Sets the GeometryId of the SketchGeometryExtension of the geometries with the provided GeoIds
        Expects a list of pairs (GeoId, id)
        """

    def getGeometryId():
        """
        Gets the GeometryId of the SketchGeometryExtension of the geometry with the provided GeoId
        """
        ...



# SPDX-License-Identifier: LGPL-2.1-or-later

from typing import Any, Final

from Base.BaseClass import BaseClass
from Base.Metadata import export

@export(
    Include="Mod/CAM/App/Area.h",
    Namespace="Path",
    Constructor=True,
    Delete=True,
)
class Area(BaseClass):
    """
    FreeCAD python wrapper of libarea

    Path.Area(key=value ...)

    The constructor accepts the same parameters as setParams(...) to configure the object
    All arguments are optional.

    Author: Zheng, Lei (realthunder.dev@gmail.com)
    License: LGPL-2.1-or-later
    """

    def add(self, **kwargs) -> Any:
        """"""
        ...

    def setPlane(self) -> Any:
        """setPlane(shape): Set the working plane.

        The supplied shape does not need to be planar. Area will try to find planar
        sub-shape (face, wire or edge). If more than one planar sub-shape is found, it
        will prefer the top plane parallel to XY0 plane. If no working plane are set,
        Area will try to find a working plane from the added children shape using the
        same algorithm"""
        ...

    def getShape(self, **kwargs) -> Any:
        """getShape(index=-1,rebuild=False): Return the resulting shape


        * index (-1): the index of the section. -1 means all sections. No effect on planar shape.


        * rebuild: clean the internal cache and rebuild"""
        ...

    def makeOffset(self, **kwargs) -> Any:
        """Make an offset of the shape."""
        ...

    def makePocket(self, **kwargs) -> Any:
        """Generate pocket toolpath of the shape."""
        ...

    def makeSections(self, **kwargs) -> Any:
        """Make a list of area holding the sectioned children shapes on given heights."""
        ...

    def getClearedArea(self) -> Any:
        """Gets the area cleared when a tool of the specified diameter follows the gcode represented in the path, ignoring cleared space above zmax and path segments that don't affect space within the x/y space of bbox."""
        ...

    def getRestArea(self) -> Any:
        """Rest machining: Gets the area left to be machined, assuming some of this area has already been cleared by previous tool paths."""
        ...

    def toTopoShape(self) -> Any:
        """Convert the Area object to a TopoShape."""
        ...

    def setParams(self, **kwargs) -> Any:
        """Set algorithm parameters."""
        ...

    def setDefaultParams(self, **kwargs) -> Any:
        """Static method to set the default parameters of all following Path.Area, plus the following additional parameters."""
        ...

    def getDefaultParams(self) -> Any:
        """Static method to return the current default parameters."""
        ...

    def getParamsDesc(self, **kwargs) -> Any:
        """Returns a list of supported parameters and their descriptions."""
        ...

    def getParams(self) -> Any:
        """Get current algorithm parameters as a dictionary."""
        ...

    def abort(self, **kwargs) -> Any:
        """Abort the current operation."""
        ...
    Sections: Final[list]
    """List of sections in this area."""

    Workplane: Any
    """The current workplane. If no plane is set, it is derived from the added shapes."""

    Shapes: Final[list]
    """A list of tuple: [(shape,op), ...] containing the added shapes together with their operation code"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# FreeCAD init script of the Mesh module
# (c) 2004 Werner Mayer LGPL

import FreeCAD

# Append the open handler
FreeCAD.addImportType("STL Mesh (*.stl *.STL *.ast *.AST)", "Mesh")
FreeCAD.addImportType("Binary Mesh (*.bms *.BMS)", "Mesh")
FreeCAD.addImportType("Alias Mesh (*.obj *.OBJ)", "Mesh")
FreeCAD.addImportType("Object File Format Mesh (*.off *.OFF)", "Mesh")
FreeCAD.addImportType("Stanford Triangle Mesh (*.ply *.PLY)", "Mesh")
FreeCAD.addImportType("Simple Model Format (*.smf *.SMF)", "Mesh")
FreeCAD.addImportType("3D Manufacturing Format (*.3mf *.3MF)", "Mesh")

FreeCAD.addExportType("STL Mesh (*.stl *.ast)", "Mesh")
FreeCAD.addExportType("Binary Mesh (*.bms)", "Mesh")
FreeCAD.addExportType("Alias Mesh (*.obj)", "Mesh")
FreeCAD.addExportType("Object File Format Mesh (*.off)", "Mesh")
FreeCAD.addExportType("Stanford Triangle Mesh (*.ply)", "Mesh")
FreeCAD.addExportType("Additive Manufacturing Format (*.amf)", "Mesh")
FreeCAD.addExportType("Simple Model Format (*.smf)", "Mesh")
FreeCAD.addExportType("3D Manufacturing Format (*.3mf)", "Mesh")

FreeCAD.__unit_test__ += ["MeshTestsApp"]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009 Yorik van Havre <yorik@uncreated.net>                        *
# *   (c) 2010 Ken Cline <cline@frii.com>                                   *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create BSpline objects.

See https://en.wikipedia.org/wiki/B-spline
"""
## @package gui_splines
# \ingroup draftguitools
# \brief Provides GUI tools to create BSpline objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import draftutils.utils as utils
import draftutils.todo as todo
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftguitools.gui_lines as gui_lines
import draftguitools.gui_trackers as trackers

from draftutils.messages import _err, _toolmsg
from draftutils.translate import translate


class BSpline(gui_lines.Line):
    """Gui command for the BSpline tool."""

    def __init__(self):
        super(BSpline, self).__init__(mode="wire")

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_BSpline",
            "Accel": "B, S",
            "MenuText": QT_TRANSLATE_NOOP("Draft_BSpline", "B-Spline"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_BSpline", "Creates a multiple-point B-spline"),
        }

    def Activated(self):
        """Execute when the command is called.

        Activate the specific BSpline tracker.
        """
        super(BSpline, self).Activated(
            name="Bspline", icon="Draft_BSpline", task_title=translate("draft", "B-Spline")
        )
        if self.doc:
            self.bsplinetrack = trackers.bsplineTracker()

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view
        by the `Activated` method of the parent class.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
            return
        if not self.ui.mouse:
            return
        if arg["Type"] == "SoLocation2Event":  # mouse movement detection
            self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg, noTracker=True)
            self.bsplinetrack.update(self.node + [self.point])
            gui_tool_utils.redraw3DView()
            return
        if arg["Type"] != "SoMouseButtonEvent":
            return
        if arg["State"] == "UP":
            self.obj.ViewObject.Selectable = True
            return
        if arg["State"] == "DOWN" and arg["Button"] == "BUTTON1":
            # Stop self.obj from being selected to avoid its display in the tree:
            self.obj.ViewObject.Selectable = False
            if arg["Position"] == self.pos:
                self.finish(cont=None)
                return
            if (not self.node) and (not self.support):
                gui_tool_utils.getSupport(arg)
                self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg, noTracker=True)
            if self.point:
                self.ui.redraw()
                self.pos = arg["Position"]
                self.node.append(self.point)
                self.drawUpdate(self.point)
                if self.mode == "line" and len(self.node) == 2:
                    self.finish(cont=None, closed=False)
                if len(self.node) > 2:
                    # DNC: allows one to close the curve
                    # by placing ends close to each other
                    # with tol = Draft tolerance
                    # old code has been to insensitive
                    if (self.point - self.node[0]).Length < utils.tolerance():
                        self.undolast()
                        self.finish(cont=None, closed=True)

    def undolast(self):
        """Undo last line segment."""
        import Part

        if len(self.node) > 1:
            self.node.pop()
            self.bsplinetrack.update(self.node)
            spline = Part.BSplineCurve()
            spline.interpolate(self.node, False)
            self.obj.Shape = spline.toShape()
            self.update_hints()

    def drawUpdate(self, point):
        """Draw and update to the spline."""
        import Part

        if len(self.node) == 1:
            self.bsplinetrack.on()
            if self.planetrack:
                self.planetrack.set(self.node[0])
            _toolmsg(translate("draft", "Pick next point"))
        else:
            spline = Part.BSplineCurve()
            spline.interpolate(self.node, False)
            self.obj.Shape = spline.toShape()
            _toolmsg(translate("draft", "Pick next point"))
        self.update_hints()

    def finish(self, cont=False, closed=False):
        """Terminate the operation and close the spline if asked.

        Parameters
        ----------
        cont: bool or None, optional
            Restart (continue) the command if `True`, or if `None` and
            `ui.continueMode` is `True`.
        closed: bool, optional
            Close the spline if `True`.
        """
        self.end_callbacks(self.call)
        if self.ui:
            self.bsplinetrack.finalize()
        if self.obj:
            # Remove temporary object, if any
            old = self.obj.Name
            todo.ToDo.delay(self.doc.removeObject, old)
        if len(self.node) > 1:
            # The command to run is built as a series of text strings
            # to be committed through the `draftutils.todo.ToDo` class.
            try:
                rot, sup, pts, fil = self.getStrings()
                Gui.addModule("Draft")

                _cmd = "Draft.make_bspline"
                _cmd += "("
                _cmd += "points, "
                _cmd += "closed=" + str(closed) + ", "
                _cmd += "face=" + fil + ", "
                _cmd += "support=" + sup
                _cmd += ")"
                _cmd_list = [
                    "points = " + pts,
                    "spline = " + _cmd,
                    "Draft.autogroup(spline)",
                    "FreeCAD.ActiveDocument.recompute()",
                ]
                self.commit(translate("draft", "Create B-Spline"), _cmd_list)
            except Exception:
                _err("Draft: error delaying commit")

        # `Creator` is the grandfather class, the parent of `Line`;
        # we need to call it to perform final cleanup tasks.
        #
        # Calling it directly like this is a bit messy; maybe we need
        # another method that performs cleanup (superfinish)
        # that is not re-implemented by any of the child classes.
        gui_base_original.Creator.finish(self)
        if cont or (cont is None and self.ui and self.ui.continueMode):
            self.Activated()


Gui.addCommand("Draft_BSpline", BSpline())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import Path
from enum import Enum

__title__ = "Toolchange toolpath Generator"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Generates the rotation toolpath"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class SpindleDirection(Enum):
    OFF = "OFF"
    CW = "M3"
    CCW = "M4"


def generate(toolnumber, toollabel, spindlespeed=0, spindledirection=SpindleDirection.OFF):
    """
    Generates Gcode for a simple toolchange.

    """

    Path.Log.track(
        f"toolnumber:{toolnumber} toollabel: {toollabel} spindlespeed:{spindlespeed} spindledirection: {spindledirection}"
    )

    if spindledirection is not SpindleDirection.OFF and spindlespeed == 0:
        spindledirection = SpindleDirection.OFF
        # raise ValueError("Turning on spindle with zero speed is invalid")

    if spindlespeed < 0:
        raise ValueError("Spindle speed must be a positive value")

    commands = []

    commands.append(Path.Command(f"({toollabel})"))
    commands.append(Path.Command("M6", {"T": int(toolnumber)}))

    if spindledirection is SpindleDirection.OFF:
        return commands
    else:
        commands.append(Path.Command(spindledirection.value, {"S": spindlespeed}))

    Path.Log.track(commands)
    return commands


def generateSubstitute(newTC, oldTC=None):
    """
    The specific commands to emit, depend on the state of the machine.
    For example, the toolnumber may not change, only the spindle speed.
    This routine will generate a list of commands to substitute for a TC
    object to be handed to the postprocessor.
    It will contain only the commands needed
    """

    if oldTC is None:
        return newTC.Path.Commands

    if newTC.ToolNumber != oldTC.ToolNumber:  # Full toolchange
        return newTC.Path.Commands

    if (newTC.SpindleSpeed != oldTC.SpindleSpeed) or (newTC.SpindleDir != oldTC.SpindleDir):
        if newTC.SpindleDir == "Forward":
            sd = SpindleDirection.CW
        elif newTC.SpindleDir == "Reverse":
            sd = SpindleDirection.CCW
        else:
            sd = SpindleDirection.OFF

        return [Path.Command(sd.value, {"S": newTC.SpindleSpeed})]

    return []



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path.Base.PropertyBag as PathPropertyBag
import CAMTests.PathTestUtils as PathTestUtils


def as_group_list(groups):
    """Normalize CustomPropertyGroups to a list of strings."""
    if groups is None:
        return []
    if isinstance(groups, (list, tuple)):
        return list(groups)
    if isinstance(groups, str):
        return [groups]
    try:
        return list(groups)
    except Exception:
        return [str(groups)]


class TestPathPropertyBag(PathTestUtils.PathTestBase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("test-property-bag")

    def tearDown(self):
        FreeCAD.closeDocument(self.doc.Name)

    def test00(self):
        """basic PropertyBag creation and access test"""
        bag = PathPropertyBag.Create()
        self.assertTrue(hasattr(bag, "Proxy"))
        self.assertEqual(bag.Proxy.getCustomProperties(), [])
        self.assertEqual(as_group_list(bag.CustomPropertyGroups), [])

    def test01(self):
        """adding properties to a PropertyBag is tracked properly"""
        bag = PathPropertyBag.Create()
        proxy = bag.Proxy
        proxy.addCustomProperty("App::PropertyString", "Title", "Address", "Some description")
        self.assertTrue(hasattr(bag, "Title"))
        bag.Title = "Madame"
        self.assertEqual(bag.Title, "Madame")
        self.assertEqual(bag.Proxy.getCustomProperties(), ["Title"])
        self.assertEqual(as_group_list(bag.CustomPropertyGroups), ["Address"])

    def test02(self):
        """refreshCustomPropertyGroups deletes empty groups"""
        bag = PathPropertyBag.Create()
        proxy = bag.Proxy
        proxy.addCustomProperty("App::PropertyString", "Title", "Address", "Some description")
        bag.Title = "Madame"
        bag.removeProperty("Title")
        proxy.refreshCustomPropertyGroups()
        self.assertEqual(bag.Proxy.getCustomProperties(), [])
        self.assertEqual(as_group_list(bag.CustomPropertyGroups), [])

    def test03(self):
        """refreshCustomPropertyGroups does not delete non-empty groups"""
        bag = PathPropertyBag.Create()
        proxy = bag.Proxy
        proxy.addCustomProperty("App::PropertyString", "Title", "Address", "Some description")
        proxy.addCustomProperty("App::PropertyString", "Gender", "Attributes")
        bag.Title = "Madame"
        bag.Gender = "Female"
        bag.removeProperty("Gender")
        proxy.refreshCustomPropertyGroups()
        self.assertEqual(bag.Proxy.getCustomProperties(), ["Title"])
        self.assertEqual(as_group_list(bag.CustomPropertyGroups), ["Address"])



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the base Draft object.

Many viewprovider classes may inherit this class in order to have
the same basic behavior."""
## @package view_base
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the base Draft object.

## \addtogroup draftviewproviders
# @{
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.translate import translate

if App.GuiUp:
    from pivy import coin
    import FreeCADGui as Gui
    import Draft_rc

    # The module is used to prevent complaints from code checkers (flake8)
    bool(Draft_rc.__name__)


class ViewProviderDraft(object):
    """The base class for Draft view providers.

    Parameters
    ----------
    vobj : a base C++ view provider
        The view provider of the scripted object (`obj.ViewObject`),
        which commonly may be of types `PartGui::ViewProvider2DObjectPython`,
        `PartGui::ViewProviderPython`, or `Gui::ViewProviderFeaturePython`.

        A basic view provider is instantiated during the creation
        of the base C++ object, for example,
        `Part::Part2DObjectPython`, `Part::FeaturePython`,
        or `App::FeaturePython`.

            >>> obj = App.ActiveDocument.addObject('Part::Part2DObjectPython')
            >>> vobj = obj.ViewObject
            >>> ViewProviderDraft(vobj)

        This view provider class instance is stored in the `Proxy` attribute
        of the base view provider.
        ::
            vobj.Proxy = self

    Attributes
    ----------
    Object : the base C++ object
        The scripted document object that is associated
        with this view provider, which commonly may be of types
        `Part::Part2DObjectPython`, `Part::FeaturePython`,
        or `App::FeaturePython`.

    texture : coin.SoTexture2
        A texture that could be added to this object.

    texcoords : coin.SoTextureCoordinatePlane
        The coordinates defining a plane to use for aligning the texture.

    These class attributes are accessible through the `Proxy` object:
    `vobj.Proxy.Object`, `vobj.Proxy.texture`, etc.
    """

    def __init__(self, vobj):
        self.Object = vobj.Object
        self.texture = None
        self.texcoords = None

        self._set_properties(vobj)
        # This class is assigned to the Proxy attribute
        vobj.Proxy = self

    def _set_properties(self, vobj):
        """Set the properties of objects if they don't exist."""
        if not hasattr(vobj, "Pattern"):
            vobj.addProperty(
                "App::PropertyEnumeration",
                "Pattern",
                "Draft",
                QT_TRANSLATE_NOOP("App::Property", "Defines an SVG pattern."),
                locked=True,
            )
            patterns = list(utils.svg_patterns())
            patterns.sort()
            vobj.Pattern = ["None"] + patterns

        if not hasattr(vobj, "PatternSize"):
            vobj.addProperty(
                "App::PropertyFloat",
                "PatternSize",
                "Draft",
                QT_TRANSLATE_NOOP("App::Property", "Defines the size of the SVG pattern."),
                locked=True,
            )
            vobj.PatternSize = params.get_param("HatchPatternSize")

    def dumps(self):
        """Return a tuple of all serializable objects or None.

        When saving the document this view provider object gets stored
        using Python's `json` module.

        Since we have some un-serializable objects (Coin objects) in here
        we must define this method to return a tuple of all serializable
        objects or `None`.

        Override this method to define the serializable objects to return.

        By default it returns `None`.

        Returns
        -------
        None
        """
        return None

    def loads(self, state):
        """Set some internal properties for all restored objects.

        When a document is restored this method is used to set some properties
        for the objects stored with `json`.

        Override this method to define the properties to change for the
        restored serialized objects.

        By default no objects were serialized with `dumps`,
        so nothing needs to be done here, and it returns `None`.

        Parameters
        ----------
        state : state
            A serialized object.

        Returns
        -------
        None
        """
        return None

    def attach(self, vobj):
        """Set up the scene sub-graph of the view provider.

        This method should always be defined, even if it does nothing.

        Override this method to set up a custom scene.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.
        """
        self.texture = None
        self.texcoords = None
        self.Object = vobj.Object
        self.onChanged(vobj, "Pattern")
        return

    def updateData(self, obj, prop):
        """Run when an object property is changed.

        Override this method to handle the behavior of the view provider
        depending on changes that occur to the real object's properties.

        By default, no property is tested, and it does nothing.

        Parameters
        ----------
        obj : the base C++ object
            The scripted document object that is associated
            with this view provider, which commonly may be of types
            `Part::Part2DObjectPython`, `Part::FeaturePython`,
            or `App::FeaturePython`.

        prop : str
            Name of the property that was modified.
        """
        return

    def getDisplayModes(self, vobj):
        """Return a list of display modes.

        Override this method to return a list of strings with
        display mode styles, such as `'Flat Lines'`, `'Shaded'`,
        `'Wireframe'`, `'Points'`.

        By default it returns an empty list.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        Returns
        -------
        list
            Empty list `[ ]`
        """
        modes = []
        return modes

    def getDefaultDisplayMode(self):
        """Return the default mode defined in getDisplayModes.

        Override this method to return a string with the default display mode.

        By default it returns `'Flat Lines'`.

        Returns
        -------
        str
            `'Flat Lines'`

        """
        return "Flat Lines"

    def setDisplayMode(self, mode):
        """Map the modes defined in attach with those in getDisplayModes.

        This method is optional.

        By default since they have the same names nothing needs to be done,
        and it just returns the input `mode`.

        Parameters
        ----------
        str
            A string defining a display mode such as
            `'Flat Lines'`, `'Shaded'`, `'Wireframe'`, `'Points'`.
        """
        return mode

    def onChanged(self, vobj, prop):
        """Run when a view property is changed.

        Override this method to handle the behavior
        of the view provider depending on changes that occur to its properties
        such as line color, line width, point color, point size,
        draw style, shape color, transparency, and others.

        This method  updates the texture and pattern if
        the properties `TextureImage`, `Pattern`, `ShapeAppearance`,
        and `PatternSize` change.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        prop : str
            Name of the property that was modified.
        """
        # treatment of patterns and image textures
        if prop in ("TextureImage", "Pattern", "ShapeAppearance"):
            if hasattr(self.Object, "Shape") and self.Object.Shape.Faces:
                path = None
                if hasattr(vobj, "TextureImage"):
                    if vobj.TextureImage:
                        path = vobj.TextureImage
                if not path:
                    if hasattr(vobj, "Pattern"):
                        if str(vobj.Pattern) in utils.svg_patterns():
                            path = utils.svg_patterns()[vobj.Pattern][1]
                        else:
                            path = "None"
                if path and vobj.RootNode:
                    switch = gui_utils.find_coin_node(vobj.RootNode, coin.SoSwitch)
                    if switch is not None and switch.getChildren().getLength() > 0:
                        flat_lines_node = switch.getChild(0)
                        # Get the Shaded sub-node of the Flat Lines node. The name in
                        # Std_SceneInspector is "Shaded", but the actual name is "FlatRoot")
                        shaded_node = gui_utils.find_coin_node_by_name(flat_lines_node, "FlatRoot")
                        if shaded_node is not None:
                            i = QtCore.QFileInfo(path)
                            if self.texture:
                                shaded_node.removeChild(self.texture)
                                self.texture = None
                            if self.texcoords:
                                shaded_node.removeChild(self.texcoords)
                                self.texcoords = None
                            if i.exists():
                                size = None
                                if ".SVG" in path.upper():
                                    size = params.get_param("HatchPatternResolution")
                                    if not size:
                                        size = 128
                                im = gui_utils.load_texture(path, size)
                                if im:
                                    self.texture = coin.SoTexture2()
                                    self.texture.image = im
                                    shaded_node.insertChild(self.texture, 1)
                                    if size:
                                        s = 1
                                        if hasattr(vobj, "PatternSize"):
                                            if vobj.PatternSize:
                                                s = vobj.PatternSize
                                        self.texcoords = coin.SoTextureCoordinatePlane()
                                        self.texcoords.directionS.setValue(s, 0, 0)
                                        self.texcoords.directionT.setValue(0, s, 0)
                                        shaded_node.insertChild(self.texcoords, 2)
        elif prop == "PatternSize":
            if hasattr(self, "texcoords"):
                if self.texcoords:
                    s = 1
                    if vobj.PatternSize:
                        s = vobj.PatternSize
                    vS = App.Vector(self.texcoords.directionS.getValue().getValue())
                    vT = App.Vector(self.texcoords.directionT.getValue().getValue())
                    vS.Length = s
                    vT.Length = s
                    self.texcoords.directionS.setValue(vS.x, vS.y, vS.z)
                    self.texcoords.directionT.setValue(vT.x, vT.y, vT.z)
        return

    def _update_pattern_size(self, vobj):
        """Update the pattern size. Helper method in onChanged."""
        if hasattr(self, "texcoords") and self.texcoords:
            s = 1
            if vobj.PatternSize:
                s = vobj.PatternSize
            vS = App.Vector(self.texcoords.directionS.getValue().getValue())
            vT = App.Vector(self.texcoords.directionT.getValue().getValue())
            vS.Length = s
            vT.Length = s
            self.texcoords.directionS.setValue(vS.x, vS.y, vS.z)
            self.texcoords.directionT.setValue(vT.x, vT.y, vT.z)

    def execute(self, vobj):
        """Run when the object is created or recomputed.

        Override this method to produce effects when the object
        is newly created, and whenever the document is recomputed.

        By default it does nothing.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.
        """
        return

    def setEdit(self, vobj, mode):
        """Enter the edit mode of the object.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        mode : int
            0, 1, 2 or 3. See the `Std_UserEditMode` command.

        Returns
        -------
        bool or None
            `False`: NOT USED HERE.
                     Do not enter edit mode (unsetEdit is not called).
            `True` : Handled edit mode.
                     Handled modes are 0 and 3, and only for certain objects.
                     The unsetEdit function should then also return `True`.
            `None` : Unhandled edit mode.
                     Handling is left to Part::FeaturePython code.
                     The unsetEdit function should then also return `None`.
        """
        if mode == 1 or mode == 2:
            return None

        # Fillet, Point, Shape2DView and PanelCut objects rely on a doubleClicked
        # function which takes precedence over all double-click edit modes. This
        # is a workaround as calling Gui.runCommand("Std_TransformManip") from
        # setEdit does not work properly. The object then seems to be put into
        # edit mode twice (?) and Esc will not cancel the command.

        # Facebinder, ShapeString, PanelSheet and Profile objects have their own
        # setEdit and unsetEdit.

        if utils.get_type(vobj.Object) in (
            "Wire",
            "Circle",
            "Ellipse",
            "Rectangle",
            "Polygon",
            "BSpline",
            "BezCurve",
        ):
            if not "Draft_Edit" in Gui.listCommands():
                self.wb_before_edit = Gui.activeWorkbench()
                Gui.activateWorkbench("DraftWorkbench")
            Gui.runCommand("Draft_Edit")
            return True

        return None

    def unsetEdit(self, vobj, mode):
        """Terminate the edit mode of the object.

        See setEdit.
        """
        if mode == 1 or mode == 2:
            return None

        if utils.get_type(vobj.Object) in (
            "Wire",
            "Circle",
            "Ellipse",
            "Rectangle",
            "Polygon",
            "BSpline",
            "BezCurve",
        ):
            if hasattr(App, "activeDraftCommand") and App.activeDraftCommand:
                App.activeDraftCommand.finish()
            Gui.Control.closeDialog()
            if hasattr(self, "wb_before_edit"):
                Gui.activateWorkbench(self.wb_before_edit.name())
                delattr(self, "wb_before_edit")
            return True

        return None

    def setupContextMenu(self, vobj, menu):
        tp = utils.get_type(self.Object)

        if tp in (
            "Wire",
            "Circle",
            "Ellipse",
            "Rectangle",
            "Polygon",
            "BSpline",
            "BezCurve",
            "Facebinder",
            "ShapeString",
            "PanelSheet",
            "Profile",
        ):
            action_edit = QtGui.QAction(translate("draft", "Edit"), menu)
            QtCore.QObject.connect(action_edit, QtCore.SIGNAL("triggered()"), self.edit)
            menu.addAction(action_edit)

        if tp == "Wire":
            action_flatten = QtGui.QAction(translate("draft", "Flatten"), menu)
            QtCore.QObject.connect(
                action_flatten, QtCore.SIGNAL("triggered()"), self.flatten
            )  # The flatten function is defined in view_wire.py.
            menu.addAction(action_flatten)

        # The default Part::FeaturePython context menu contains a `Set colors`
        # option. This option makes no sense for objects without a face or that
        # can only have a single face. In those cases we override this menu and
        # have to add our own `Transform` item.
        # To override the default menu this function must return `True`.
        if tp in (
            "Wire",
            "Circle",
            "Ellipse",
            "Rectangle",
            "Polygon",
            "BSpline",
            "BezCurve",
            "Fillet",
            "Point",
            "Shape2DView",
            "PanelCut",
            "PanelSheet",
            "Profile",
        ):
            action_transform = QtGui.QAction(
                Gui.getIcon("Std_TransformManip.svg"),
                translate("Command", "Transform"),  # Context `Command` instead of `draft`.
                menu,
            )
            QtCore.QObject.connect(action_transform, QtCore.SIGNAL("triggered()"), self.transform)
            menu.addAction(action_transform)
            return True

    def edit(self):
        Gui.ActiveDocument.setEdit(self.Object, 0)

    def transform(self):
        Gui.ActiveDocument.setEdit(self.Object, 1)

    def getIcon(self):
        """Return the path to the icon used by the view provider.

        The path can be a full path in the system, or a relative path
        inside the compiled resource file.
        It can also be a string that defines the icon in XPM format.

        Override this method to provide a specific icon
        for the object in the tree view.

        By default it returns the path to the `Draft_Draft.svg` icon.

        Returns
        -------
        str
            `':/icons/Draft_Draft.svg'`
        """
        tp = utils.get_type(self.Object)
        if tp in ("Line", "Wire", "Polyline"):
            return ":/icons/Draft_N-Linear.svg"
        if tp in ("Rectangle", "Polygon"):
            return ":/icons/Draft_N-Polygon.svg"
        if tp in ("Circle", "Ellipse", "BSpline", "BezCurve", "Fillet"):
            return ":/icons/Draft_N-Curve.svg"
        if tp in ("ShapeString"):
            return ":/icons/Draft_ShapeString_tree.svg"
        if hasattr(self.Object, "AutoUpdate") and not self.Object.AutoUpdate:
            import TechDrawGui

            return ":/icons/TechDraw_TreePageUnsync.svg"
        return ":/icons/Draft_Draft.svg"

    def claimChildren(self):
        """Return objects that will be placed under it in the tree view.

        Override this method to return a list with objects
        that will appear under this object in the tree view.
        That is, this object becomes the `parent`,
        and all those under it are the `children`.

        By default the returned list is composed of objects from
        `Object.Base`, `Object.Objects`, `Object.Components`,
        and `Object.Group`, if they exist.

        Returns
        -------
        list
            List of objects.
        """
        objs = []
        if hasattr(self.Object, "Base"):
            objs.append(self.Object.Base)
        if hasattr(self.Object, "Objects"):
            objs.extend(self.Object.Objects)
        if hasattr(self.Object, "Components"):
            objs.extend(self.Object.Components)
        if hasattr(self.Object, "Group"):
            objs.extend(self.Object.Group)
        return objs


# Alias for compatibility with v0.18 and earlier
_ViewProviderDraft = ViewProviderDraft


class ViewProviderDraftAlt(ViewProviderDraft):
    """A view provider that doesn't absorb its base object in the tree view.

    The `claimChildren` method is overridden to return an empty list.

    The `doubleClicked` method is defined.

    Only used by the `Shape2DView` object.
    """

    def __init__(self, vobj):
        super(ViewProviderDraftAlt, self).__init__(vobj)

    def doubleClicked(self, vobj):
        # See setEdit in ViewProviderDraft.
        Gui.runCommand("Std_TransformManip")
        return True

    def claimChildren(self):
        return []


# Alias for compatibility with v0.18 and earlier
_ViewProviderDraftAlt = ViewProviderDraftAlt


class ViewProviderDraftPart(ViewProviderDraftAlt):
    """A view provider that displays a Part icon instead of a Draft icon.

    The `getIcon` method is overridden to provide `Part_3D_object.svg`.

    Only used by the `Block` object.
    """

    def __init__(self, vobj):
        super(ViewProviderDraftPart, self).__init__(vobj)

    def getIcon(self):
        return ":/icons/Part_3D_object.svg"

    def setEdit(self, vobj, mode):
        return None

    def unsetEdit(self, vobj, mode):
        return None


# Alias for compatibility with v0.18 and earlier
_ViewProviderDraftPart = ViewProviderDraftPart

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.MachineState as PathMachineState
import Part

__title__ = "Feed Rate Helper Utility"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Helper for adding Feed Rate to Path Commands"

"""
TODO:  This needs to be able to handle feedrates for axes other than X,Y,Z
"""

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def setFeedRate(commandlist, ToolController):
    """Set the appropriate feed rate for a list of Path commands using the information from a Tool Controller

    Every motion command in the list will have a feed rate parameter added or overwritten based
    on the information stored in the tool controller. If a motion is a plunge (vertical) motion, the
    VertFeed value will be used, otherwise the HorizFeed value will be used instead."""

    def _isVertical(currentposition, command):
        x = command.Parameters["X"] if "X" in command.Parameters else currentposition.x
        y = command.Parameters["Y"] if "Y" in command.Parameters else currentposition.y
        z = command.Parameters["Z"] if "Z" in command.Parameters else currentposition.z
        endpoint = FreeCAD.Vector(x, y, z)
        if Path.Geom.pointsCoincide(currentposition, endpoint):
            return True
        return Path.Geom.isVertical(Part.makeLine(currentposition, endpoint))

    machine = PathMachineState.MachineState()

    for command in commandlist:
        if command.Name not in Path.Geom.CmdMoveAll:
            continue

        if _isVertical(machine.getPosition(), command):
            rate = (
                ToolController.VertRapid.Value
                if command.Name in Path.Geom.CmdMoveRapid
                else ToolController.VertFeed.Value
            )
        else:
            rate = (
                ToolController.HorizRapid.Value
                if command.Name in Path.Geom.CmdMoveRapid
                else ToolController.HorizFeed.Value
            )

        params = command.Parameters
        params["F"] = rate
        command.Parameters = params

        machine.addCommand(command)

    return commandlist
